# -*- coding: utf-8 -*-
"""This module contains functions for model agnostic interpretation methods. 
"""

import numpy as np
from pailab.ml_repo.repo_objects import RepoObject, RawData, RepoInfoKey
from pailab.tools.tools import ml_cache
from pailab.ml_repo.repo_store import RepoStore  # pylint: disable=E0401
from pailab.ml_repo.repo import MLObjectType


has_sklearn = True
try:
    from sklearn.cluster import KMeans
    from sklearn.metrics.pairwise import pairwise_kernels
    from sklearn import preprocessing
except ImportError:
    import warnings
    warnings.warn('No sklearn installed, some functions of this submodule may not be usable.')
    has_sklearn = False


@ml_cache
def _compute_ice(data, model_eval_function, model,
                 y_coordinate, x_coordinate,
                 x_values, start_index = 0, end_index = -1, scale=''):
    """Independent conditional expectation plot

    Args:
        data ([type]): [description]
        model_eval_function ([type]): [description]
        model ([type]): [description]
        direction ([type]): [description]
        y_coordinate ([type]): [description]
        x_coordinate ([type]): [description]
        x_values ([type]): [description]
        scale (str or int, optional): String defining the scaling for the functions before functional clustering is applied. Scaling is perfomred by
                                dividing the vector of the y-values of the ICE by the respective vector norm defined by scaling.
                                The scaling must be one of numpy's valid strings for linalg.norm's ord parameter. If string is empty, no scaling will be applied.
                                Defaults to ''. 
    Returns:
        [type]: [description]
    """
    x_data = data.x_data[start_index:end_index, :]
    
    # compute input for evaluation
    shape = (x_data.shape[0], len(x_values),)  # x_data.shape[1:]
    _x_data = np.empty(shape=(len(x_values), ) + x_data.shape[1:])
    eval_data = RawData(_x_data, [str(i) for i in range(x_data.shape[1])])
    result = np.empty(shape)
    eval_f = model_eval_function.create()
    nom = 1.0
    if isinstance(scale, str):
        use_scaling = False
    else:
        use_scaling = True

    for i in range(x_data.shape[0]):
        for j in range(len(x_values)):
            eval_data.x_data[j] = x_data[i]
            eval_data.x_data[j, x_coordinate] = x_values[j]
        tmp = eval_f(model, eval_data)
        if len(tmp.shape) > 1:
            y = tmp[:, y_coordinate]
        elif y_coordinate == 0:
            y = tmp[:]
        else:
            raise Exception('Evaluation data is just an array but y_coordinate > 0.')
        if use_scaling:
            nom = 1.0 / max(np.linalg.norm(y, ord = scale), 1e-10)
        result[i] = nom * y
    return result


def _get_model_eval(model, model_version, ml_repo, model_label=None):
    """Return model and evaluation function for given model and model version

    Args:
        model (str): The name of the (calibrated) model.
        model_version (str): The version of the calibrated model.
        ml_repo (MLRepo): The ml repo used to retrieve the model.
        model_label (str): A label defining the model to investigate (instead of defining the model).
    """

    model_version_ = model_version
    if model_label is not None:
        label = ml_repo.get(model_label)
        model = label.name
        model_version = label.version

    if isinstance(model, str):
        if len(model.split('/')) == 1:
            model_ = model + '/model'
        else:
            model_ = model
        result_model = ml_repo.get(model_, model_version_)
    else:
        result_model = model

    # now the eval function
    model_definition_name = result_model.repo_info.name.split('/')[0]
    model_definition = ml_repo.get(
        model_definition_name, result_model.repo_info.modification_info[model_definition_name])
    result_eval_f = ml_repo.get(model_definition.eval_function)
    return result_model, result_eval_f


class ICE_Results:
    def __init__(self):
        """Class holding the results of the compute_ice function.

        Attributes:
            ice (numpy matrix): Each row contains the ICE values at each datapoint.
            x_values (list): List of the used x values for the ICE projection.
            x_coord_name (str): Name of x-coordinate used for ICE.
            y_coord_name (str): Name of y-coordinate used for ICE.
            labels (numpy array): If functional clustering of ICE curves has been performed, it contains the cluster label for each datapoint.
            cluster_centers (numpy matriy): Each row contains one cluster center of the functional clustering.
            distance_to_clusters (numpy matrix): Each row contains the distance of the respective ICE graph to each of the clusters.
            data_name (str): Name of input data  used for ICE curves.
            start_index (int): Start index of input data.
            model (str): Name of model used for the ICE curves.
            model_version (str): Version of model used for the ICE curves.
            data_name (str): Name of underlying data.
            data_version (str): Version of underlying data.
        """
        self.ice = None
        self.x_values = None
        self.x_coord_name = ''
        self.y_coord_name = ''
        self.labels = None
        self.cluster_centers = None
        self.distance_to_clusters = None
        self.data_name = ''
        self.data_version = ''
        self.start_index = 0
        self.model = ''
        self.model_version = ''

    def _validate_for_comparison(self, ice_results_2):
        if self.data_name != ice_results_2.data_name:
            raise Exception('Cannot compare two ice results on different data, ice_results_1 : ' + self.data_name + ', ice_results_2: ' + ice_results_2.data_name)
        if self.data_version != ice_results_2.data_version:
            raise Exception('Cannot compare two ice results on different data, ice_results_1 : ' + self.data_version + ', ice_results_2: ' + ice_results_2.data_version)
        if self.start_index != ice_results_2.start_index:
            raise Exception('Cannot compare two ice results with different start_index, ice_results_1 : ' + str(self.start_index) + ', ice_results_2: ' + str(ice_results_2.start_index))
    
    def compute_cluster_average(self, ice_results_2,):
        """[summary]
        
        Args:
            ice_results_2 ([type]): [description]

        Returns:
            numpy matrix: Matrix containing the average of values from ice_results_2 over the different clusters from this result. 
        """
        self._validate_for_comparison(ice_results_2)
        if self.cluster_centers is None:
            raise Exception("No clusters have been computed yet.")
        result = np.empty(self.cluster_centers.shape)
        
        for i in range(self.cluster_centers.shape[0]):
            tmp = ice_results_2.ice[self.labels == i]
            result[i] =  np.mean(tmp, axis = 0)
        return result

def functional_clustering(x, scale='',
                 n_clusters=20, random_state=42):
    """Given a set of different 1D functions (represented in matrix form where each row contains the function values on a given grid), this
        method clusters those functions and tries to find typical function structures.
    
    Args:
        x (numpy matrix): Matrix containing in each row the function values at a datapoint.
        n_clusters (int, optional): Number of clusters for the functional clustering of the ICE curves. Defaults to 0.
        random_state (int, optional): [description]. Defaults to 42.
        scale (str or int, optional): String defining the scaling for the functions before functional clustering is applied. Scaling is perfomred by
                                dividing the vector of the y-values of the ICE by the respective vector norm defined by scaling.
                                The scaling must be one of numpy's valid strings for linalg.norm's ord parameter. If string is empty, no scaling will be applied.
                                Defaults to ''. 
    Returns:
        numpy vector: Vector where each value defines the corresponding cluster of the respective function (x[i] is the cluster the i-th function belongs to).
        numpy matrix: Contains in each row th distance to each cluster for the respective function.
        numpy matrix: Contains in each row a cluster centroid.
    """
    if not has_sklearn:
        raise Exception('Cannot apply functional clustering: No sklearn installed. Please either install sklearn or set n_clusters to zero.')
    k_means = KMeans(init='k-means++', n_clusters=n_clusters,
                    n_init=10, random_state=random_state)
    labels = k_means.fit_predict(x)
    # now store for each data point the distance to the respectiv cluster center
    distance_to_clusters = k_means.transform(x)
    cluster_centers =  k_means.cluster_centers_
    return labels, distance_to_clusters, cluster_centers


@ml_cache
def _compute_and_cluster_ice(data, model_eval_function, model,
                 y_coordinate, x_coordinate,
                 x_values, start_index = 0, end_index = -1, scale='',
                 n_clusters=20, random_state=42, clustering_param = None):
    """[summary]
    
    Args:
        data ([type]): [description]
        model_eval_function ([type]): [description]
        model ([type]): [description]
        y_coordinate ([type]): [description]
        x_coordinate ([type]): [description]
        x_values ([type]): [description]
        start_index (int, optional): [description]. Defaults to 0.
        end_index (int, optional): [description]. Defaults to -1.
        clustering_param (dict or None, optional): Default to None
        
    Returns:
        list of doubles: the x-values used to compute ICE
        numpy matrix: Each row contains the ice values
        numpy vector: Each entry contains the cluster of the respective point
        numpy matrix: Each row contains a cluster center
        numpy matrix: Each row contains the distances to the different clusters (i-th column contains distance to i-th cluster)
    """
    if isinstance(x_coordinate, str):
        x_coordinate = data.x_coord_names.index(x_coordinate)
    if isinstance(y_coordinate, str):
        y_coordinate = data.y_coord_names.index(y_coordinate)
    ice = _compute_ice(data, model_eval_function, model, y_coordinate, x_coordinate, x_values,
                                 start_index=start_index, end_index=end_index, scale=scale)
    labels = None
    cluster_centers = None
    distance_to_clusters = None
    if clustering_param is not None:
        labels, distance_to_clusters, cluster_centers = functional_clustering(ice, **clustering_param)
    return x_values, ice, labels, cluster_centers, distance_to_clusters

def compute_ice(ml_repo, x_values, data, model=None, model_label=None, model_version=RepoStore.LAST_VERSION,
                data_version=RepoStore.LAST_VERSION, y_coordinate=0, x_coordinate = 0,
                start_index=0, end_index=-1, cache = False,
                clustering_param = None, scale=''):
    """Compute individual conditional expectation (ice) for a given dataset and model
    
    Args:
        ml_repo (MLRepo): MLRepo used to retrieve model and data and be used in caching.
        x_values (list): List of x values for the ICE.
        data (str, DataSet, RawData): Either name of data or directly the data object which is used as basis for ICE (an ICE is computed at each datapoint of the data).
        model (str, optional): Name of model in the MLRepo for which the ICE will be computed. If None, model_label must be specified, defining the model to be used. Defaults to None.
        model_label (str, optional): Label defining the model to be used. Defaults to None.
        model_version (str, optional): Version of model to be used for ICE. Only needed if model is specified. Defaults to RepoStore.LAST_VERSION.
        data_version (str, optional): Version of data used. Defaults to RepoStore.LAST_VERSION.
        y_coordinate (int or str, optional): Defines y-coordinate (either by name or coordinate index) for which the ICE is computed. Defaults to 0.
        x_coordinate (int or str, optional): Defines x-coordinate (either by name or coordinate index) for which the ICE is computed. Defaults to 0.
        start_index (int, optional): Defines the start index of the data to be used in ICE computation (data[start_index:end_index] will be used). Defaults to 0.
        end_index (int, optional): Defines the end index of the data to be used in ICE computation (data[start_index:end_index] will be used). Defaults to -1.
        cache (bool, optional): If True, results will be cached. Defaults to False.
        clustering_param (dict or None, optional): Dictionary of parameters for method functional_clustering that is called if the parameter is not None and applies 
            functional clustering to the ICE curves.
        scale (str or int, optional): String defining the scaling for the functions before functional clustering is applied. Scaling is perfomred by
                                dividing the vector of the y-values of the ICE by the respective vector norm defined by scaling.
                                The scaling must be one of numpy's valid strings for linalg.norm's ord parameter. If string is empty, no scaling will be applied.
                                Defaults to ''. 

    Returns:
        ICE_Results: result object containing all relevant data (including functional clustering)
    """
    data_ = data
    if isinstance(data, str):
        data_ = ml_repo.get(data, data_version, full_object = True)
    if isinstance(x_coordinate, str):
        x_coordinate = data_.x_coord_names.index(x_coordinate)
    if isinstance(y_coordinate, str):
        y_coordinate = data_.y_coord_names.index(y_coordinate)
    model_, model_eval_f = _get_model_eval(
        model, model_version, ml_repo, model_label)

    cache_ = None
    if cache:
        cache_ = ml_repo

    result = ICE_Results()
    result.x_values, result.ice, result.labels, result.cluster_centers, result.distance_to_clusters = _compute_and_cluster_ice(data_, model_eval_f, model_,  y_coordinate,
                     x_coordinate=x_coordinate, x_values=x_values, 
                     start_index = start_index, end_index = end_index, cache=cache_,
                     clustering_param=clustering_param, scale = scale)
    result.x_coord_name = data_.x_coord_names[x_coordinate]
    result.y_coord_name = data_.y_coord_names[y_coordinate]
    result.data_name = data_.repo_info.name
    result.start_index = start_index
    result.model = model_.repo_info.name
    result.model_version = model_.repo_info.version
    result.data_version = data_.repo_info.version
    return result

def _compute_MMD_square(X,Y = None, metric = 'rbf', k_XX=None, k_YY=None, **kwds):
    """Compute the Maximum Mean Dicreapency (MMD) to measure how close two distributions are.
    
    Args:
        X (numpy nd-array): X points used to approximate first distribution. 
        Y (numpy nd-array): Y points used to approximate second distribution. 
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        **kwds: optional keyword parameters
            Any further parameters are passed directly to the kernel function.
    
    Returns:
        float: The squared MDM computed by [description]
    """
    if not has_sklearn:
        raise Exception('This method needs functionality form sklearn but sklearn is not installed.')
    m = float(X.shape[0])
   
    if k_XX is None:
        k_XX = pairwise_kernels(X, metric=metric, **kwds).sum()/(m**2) 
    if Y is None:
        return  k_XX  
    n = float(Y.shape[0])
    if k_YY is None:
        k_YY = pairwise_kernels(Y, metric=metric, **kwds).sum()/(n**2)
    k_XY = pairwise_kernels(X, Y=Y, metric=metric, **kwds).sum()/(m*n)
    return k_XX + k_YY - 2.0*k_XY

def _get_MMD2_X_vs_X_percentile(X, Y,  percentile = 0.1, scale = True, metric = 'rbf', **kwds):
    """This method compares the distribution of X and of values of X whose indices belong to the respective percentile of Y.
    
    Args:
        X (ndarray): X values for which the squared Maximum Mean Discrepancy to the subset of X described by indices defined as percentile of Y is computed.
        Y (ndarray): Y values used to compute indices belonging to the percentile.
        percentile (float, optional): The percentile used. Defaults to 0.1.
        scale (bool, optional): If True, the x-values are scaled to zero mean and unit variance. Defaults to True.
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        **kwds: optional keyword parameters that are passed directly to the kernel function.
    
    Returns:
        ndarray: Numpy nd array containing the respective squared MMD between the diffrent distributions.
    """
    result = np.empty([X.shape[1], Y.shape[1]])
    if scale:
        _X = preprocessing.StandardScaler().fit_transform(X)
    else:
        _X = X
    k_xx = [0]*X.shape[1]
    for i in range(X.shape[1]):
        tmp = np.reshape(_X[:,i],(X.shape[0], 1, ))
        #print('tmp: ' + str(tmp))
        k_xx[i] = _compute_MMD_square(tmp, metric = metric, **kwds)

    for i in range(Y.shape[1]):
        sorted_indices = np.argsort(Y[:,i])
        i_start = int( (1.0-percentile)*len(sorted_indices))
        indices = sorted_indices[i_start:]
        #print('x_percentile: ' + str(x_percentile))
        for j in range(X.shape[1]):
            x_percentile = np.reshape(_X[indices,j], (len(indices), 1, ) )
            #k_yy = _compute_MMD_square(x_percentile, metric = metric, **kwds)
            x = np.reshape(_X[:,j], (len(sorted_indices), 1, ))
            #print('k_xx: ' + str(k_xx[i]) + '   k_yy: ' + str(k_yy))
            result[j,i] = _compute_MMD_square(x, x_percentile, k_XX = k_xx[j], metric = metric, **kwds)
    return result


@ml_cache
def _get_MMD2_X_vs_abs_ptw_error_percentile(X, Y_pred, x_coords = None, y_coords = None, percentile = 0.1, scale = True, metric = 'rbf', **kwds):
    """This method compares the distribution of X and of values of the pointwise absolut errors between Y_target and Y_pred.
    
    Args:
        X (DataSet/RawData): X values for which the squared Maximum Mean Discrepancy to the subset of X described by indices defined as percentile of Y is computed.
        Y_pred (DataSet/RawData): predicted Y values
        percentile (float, optional): The percentile used. Defaults to 0.1.
        scale (bool, optional): If True, the x-values are scaled to zero mean and unit variance. Defaults to True.
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        **kwds: optional keyword parameters that are passed directly to the kernel function.
    
    Returns:
        ndarray: Numpy nd array containing the respective squared MMD between the diffrent distributions.
    """
    if y_coords is None:
        ptw_error = np.abs(X.y_data - Y_pred.x_data)
    else:
        ptw_error = np.abs(X.y_data[:, y_coords] - Y_pred.x_data[:,y_coords])
    if x_coords is None:
        return _get_MMD2_X_vs_X_percentile(X.x_data, ptw_error, percentile=percentile, scale=scale, metric=metric, **kwds)
    else:
        return _get_MMD2_X_vs_X_percentile(X.x_data[:, x_coords], ptw_error, percentile=percentile, scale=scale, metric=metric, **kwds)


def _compute_MMD2(X, prototypes, metric = 'rbf', **kwds):
    kernel_matrix = pairwise_kernels(X, metric=metric, **kwds)
    m = float(len(prototypes))
    n = float(kernel_matrix.shape[0])
    #print('kernel_matrix.sum()/(n**2): ' + str(kernel_matrix.sum()/(n**2)))
    #print('kernel_matrix[np.ix_(prototypes, prototypes)].sum()/(m**2): ' + str(kernel_matrix[np.ix_(prototypes, prototypes)].sum()/(m**2)))
    #print('2.0*kernel_matrix[prototypes, :].sum()/(m*n): ' + str(2.0*kernel_matrix[prototypes, :].sum()/(m*n)))
    return kernel_matrix.sum()/(n**2) + kernel_matrix[np.ix_(prototypes, prototypes)].sum()/(m**2) - 2.0*kernel_matrix[prototypes, :].sum()/(m*n)

def _compute_prototypes(X, n_prototypes, n_criticisms, metric = 'rbf', witness_penalty = 1.0, **kwds):
    """This methods computes for given datapoints prototypes and criticisms.
    
    This methods computes for given datapoints prototypes and criticisms, i.e. datapoints from th given set that are typical representatives (prototypes) and datapoints
    that are not well representatives (criticisms). Here, a simple greedy algorithm using MDM2 is used to compute the prototypes and a witness function together
    with som simple penalty are used to compute the criticisms (see e.g. C. Molnar, Interpretable Machine Learning).

    Args:
        X (numpy matrix): Set of datapoints (each row representing one datapoint)
        n_prototypes (int): Number of prototypes.
        n_criticisms (int): Number of criticisms.
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        witness_penalty (float): Penalty parameter to include some penalty to avoid to close criticisms. 
        **kwds: optional keyword parameters
            Any further parameters are passed directly to the kernel function.
    
    Raises:
        Exception: If sklearn is not installed

    Returns:
        list of int: List of indices defining the datapoints which are the resulting prototypes.
        list of int: List of indices defining the datapoints which are the resulting criticisms.
    """
    if not has_sklearn:
        raise Exception('This method needs functionality form sklearn but sklearn is not installed.')
    kernel_matrix = pairwise_kernels(X, metric=metric, **kwds)
    prototypes = []
    n = float(kernel_matrix.shape[0])
    if n_prototypes >= n:
        raise Exception('Number of prototypes must be less then number of datapoints.')
    # To compute  prototypes we minimize the Maximum Mean Discrepancy (MDM), .. math::
    #     MDM^2 = \frac{1}{m^2}\sum_{i=1}^m \sum_{j=1}^m k(z_i,z_j) - \frac{2}{mn}\sum_{i=1}^m \sum_{j=1}^n k(z_i,x_j) + \frac{1}{n^2}\sum_{i=1}^n\sum_{j=1}^n k(x_i,x_j)
    #     We are doing this using a greedy search, looking for the next prototype by simply computing which next datapoint reduces the current MDM most. 
    #     For this we compute simply
    #     the impact on the MDM if a new point x is used as a prototype. The impact is computed by .. math::
    #     \frac{2}{(m+1)^2}(k(x,x) + \sum{i=1}^mk(z_i,x)) -  \frac{2}{(m+1)n}\sum_{j=1}^n k(x,x_j)for i in range(n_prototypes):
    #max_impact  = 1.0e8
    for i in range(n_prototypes):
        m = float(len(prototypes))
        max_impact  = 1.0e8
        new_prototype = None
        for candidate in range(kernel_matrix.shape[0]):
            if candidate not in prototypes:
                impact = kernel_matrix[candidate][prototypes].sum()/((m+1)**2) - kernel_matrix[candidate,:].sum()/((m+1)*n)
                if impact < max_impact:
                    new_prototype = candidate
                    max_impact = impact    
        if new_prototype is None:
            print(str(prototypes))
            raise Exception('Cannot find a new prototype.')
        prototypes.append(new_prototype)

    m = float(len(prototypes))
    n = float(kernel_matrix.shape[0])
    criticisms = []    
    for i in range(n_criticisms):
        m_criticisms = float(len(criticisms)) 
        max_witness  = -1.0e8
        new_criticism = None
        for candidate in range(kernel_matrix.shape[0]):
            if candidate not in criticisms:
                witness = kernel_matrix[candidate].sum()/n - kernel_matrix[candidate][prototypes].sum()/m
                if m_criticisms > 0:
                    regularizer = kernel_matrix[candidate][criticisms].max()
                else:
                    regularizer = 0
                cost = abs(witness) - witness_penalty*regularizer
                if cost > max_witness:
                    max_witness = cost
                    new_criticism = candidate
        if new_criticism is None:
            raise Exception('Cannot find a new criticism.')
        criticisms.append(new_criticism)

    return prototypes, criticisms

def generate_prototypes(ml_repo, data, n_prototypes, n_criticisms, data_version = RepoStore.LAST_VERSION,
                        use_x = True, data_start_index=0, data_end_index = -1, metric = 'rbf', witness_penalty = 1.0, 
                        **kwds):
    """This methods computes for a given test/training dataset prototypes and criticisms and adds them as separate test data sets to the repository. 
    
    This methods computes for given test/training dataset prototypes and criticisms, i.e. datapoints from th given set that are typical representatives (prototypes) 
    and datapoints that are not well representatives (criticisms). Here, a simple greedy algorithm using MDM2 is used to compute the prototypes and a witness 
    function together with some simple penalty are used to compute the criticisms (see e.g. C. Molnar, Interpretable Machine Learning).

    Args:
        ml_repo (MLRepo): The repository used to retrieve data and store prototypes/criticisms.
        data (str): Name of data used for computation.
        n_prototypes (int): Number of prototypes.
        n_criticisms (int): Number of criticisms.
        data_version (str): Version of data to be used. Defaults to RepoStore.LAST_VERSION.
        use_x (bool): Flags that determine if prototypes are computed w.r.t. x or y coordinates. Defaults to True.
        data_start_index (int): Startindex of data used.
        data_end_index (int): Endindex of data used.
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        witness_penalty (float): Penalty parameter to include some penalty to avoid to close criticisms. 
        **kwds: optional keyword parameters
            Any further parameters are passed directly to the kernel function.
    
    Raises:
        Exception: If sklearn is not installed

    Returns:
        list of int: List of indices defining the datapoints which are the resulting prototypes.
        list of int: List of indices defining the datapoints which are the resulting criticisms.
    """
    if isinstance(data, str):
        data = ml_repo.get(data, data_version, full_object = True)
    if use_x:
        d = data.x_data[data_start_index:data_end_index]
        if d is None:
            raise Exception('No x-data defined.')
    else:
        d = data.y_data[data_start_index:data_end_index]
        if d is None:
            raise Exception('No y-data defined.')

    std_scale = preprocessing.StandardScaler().fit(d)
    d = std_scale.transform(d)
    prototypes, criticisms = _compute_prototypes(d, n_prototypes, n_criticisms, metric = metric, witness_penalty=witness_penalty, **kwds)
    
    result_name = data.repo_info.name+'_'+'prototypes'
    if data.y_data is None:
        result = RawData(data.x_data[data_start_index:data_end_index][prototypes], data.x_coord_names, repo_info = {RepoInfoKey.NAME: result_name,  
                        RepoInfoKey.CATEGORY: MLObjectType.TEST_DATA, 
                        RepoInfoKey.MODIFICATION_INFO: {data.repo_info.name: data.repo_info.version}
                    })
    else:
        result = RawData(data.x_data[data_start_index:data_end_index][prototypes], data.x_coord_names, 
            data.y_data[data_start_index:data_end_index][prototypes], data.y_coord_names, 
            repo_info = {RepoInfoKey.NAME: result_name,  
                        RepoInfoKey.CATEGORY: MLObjectType.TEST_DATA, 
                        RepoInfoKey.MODIFICATION_INFO: {data.repo_info.name: data.repo_info.version}
                    })
    ml_repo.add(result)

    result_name = data.repo_info.name+'_'+'criticisms'
    if data.y_data is None:
        result = RawData(data.x_data[data_start_index:data_end_index][criticisms], data.x_coord_names, repo_info = {RepoInfoKey.NAME: result_name,  
                        RepoInfoKey.CATEGORY: MLObjectType.TEST_DATA, 
                        RepoInfoKey.MODIFICATION_INFO: {data.repo_info.name: data.repo_info.version}
                    })
    else:
        result = RawData(data.x_data[data_start_index:data_end_index][criticisms], data.x_coord_names, 
            data.y_data[data_start_index:data_end_index][criticisms], data.y_coord_names, 
            repo_info = {RepoInfoKey.NAME: result_name,  
                        RepoInfoKey.CATEGORY: MLObjectType.TEST_DATA, 
                        RepoInfoKey.MODIFICATION_INFO: {data.repo_info.name: data.repo_info.version}
                    })
    ml_repo.add(result)


from ipywidgets import *
import ipywidgets as widgets
from IPython.display import display, clear_output

from pailab import MLObjectType, RepoInfoKey, FIRST_VERSION, LAST_VERSION
import pailab.tools.checker as checker
import pandas as pd

# set option so that long lines have a linebreak
pd.set_option('display.max_colwidth', -1)

beakerX = False
if beakerX:
    from beakerx import *
    from beakerx.object import beakerx
else:
    def TableDisplay(dt):
        display(dt)


class ObjectOverviewList:
    def __init__(self, ml_repo, beakerX=False):
        self._ml_repo = ml_repo
        self._categories = widgets.SelectMultiple(
            options=[k.value for k in MLObjectType])
        self._repo_info=widgets.SelectMultiple(
            options = [k.value for k in RepoInfoKey], value = ['category', 'name']
        )
        self._button_update=widgets.Button(description = 'update')
        self._button_update.on_click(self.get_overview)
        self._output=widgets.Output()
        self._input_box=widgets.HBox(
            children = [widgets.VBox(
                children=[self._categories, self._repo_info]), self._button_update, self._output]
        )

    def get_overview(self, d):
        result={}
        for info in self._repo_info.value:
            result[info]=[]

        for k in self._categories.value:
            names=self._ml_repo.get_names(k)
            for n in names:
                obj=self._ml_repo.get(n)
                for info in self._repo_info.value:
                    if isinstance(obj.repo_info[info], MLObjectType):
                        result[info].append(obj.repo_info[info].value)
                    else:
                        result[info].append(str(obj.repo_info[info]))
        with self._output:
            clear_output(wait = True)
            TableDisplay(pd.DataFrame.from_dict(result))  # , orient='index'))

        # _object_type = self.select_category.value
        # if _object_type == 'ALL':
        #     _object_type = ''
        # fields = []
        # for x in self.cb_additional_fields:
        #     if x.value:
        #         fields.append(x.description)
        # overview = repo.get_overview(_object_type, fields)
        # for version, x in overview.items():
        #     for key, value in x.items():
        #         if isinstance(value, dict):
        #             x[key] = str(value)
        # with self.out_objects_stored:
        #     clear_output(wait=True)
        #     display(pd.DataFrame.from_dict(overview, orient='index'))
        # self.out_objects_overview.selected_index = 1
        # # return pd.DataFrame.from_dict(overview,orient='index')
    # display()

    def get_widget(self):
        return self._input_box
        # self.cb_additional_fields = []
        # return self.final_widget

class ObjectView:

    def _setup_names(self, change = None):
        names=[]
        for k in self._categories.value:
            names.extend(self._ml_repo.get_names(k))
        self._names.options=names

    def __init__(self, ml_repo, beakerX = False):
        self._ml_repo=ml_repo
        self._categories=widgets.SelectMultiple(
            options = [k.value for k in MLObjectType], value = ['CALIBRATED_MODEL'])
        self._names=widgets.SelectMultiple(
            options = []
        )
        self._setup_names()
        self._categories.observe(self._setup_names, 'value')

        self._button_update=widgets.Button(description = 'show history')
        self._button_update.on_click(self.show_history)
        self._output=widgets.Output()
        self._input_box=widgets.HBox(
            children = [widgets.VBox(
                children=[self._categories, self._names]), self._button_update, self._output]
        )

    def show_history(self, d):
        result={RepoInfoKey.NAME.value: [],
             RepoInfoKey.AUTHOR.value: [],
             RepoInfoKey.VERSION.value: [],
             RepoInfoKey.COMMIT_DATE.value: []}
        for k in self._names.value:
            history = self._ml_repo.get_history(k)
            for l in history:
                for m in result.keys():
                    result[m].append(l['repo_info'][m])
        with self._output:
            clear_output(wait=True)
            TableDisplay(pd.DataFrame.from_dict(result))

    def get_widget(self):
        return self._input_box

class ConsistencyChecker:

    def _consistency_check(self):
        self._test_results = checker.Tests.run(self._ml_repo)
        self._model_results = checker.Model.run(self._ml_repo)
        self._data_results = checker.Data.run(self._ml_repo)

    def __init__(self, ml_repo, beakerX=False):
        self._ml_repo = ml_repo

        self._overview_output = widgets.Output()

        self._button_update = widgets.Button(description='update')
        self._button_update.on_click(self.show_checks)

        self._widget_main = widgets.VBox(
            children=[self._button_update, self._overview_output]
        )

    def show_checks(self, d):
        self._consistency_check()
        with self._overview_output:
            clear_output(wait=True)
            print('test issues: ' + str(len(self._test_results)) + ', model issues: ' + str(len(self._model_results))+ ', data issues: ' + str(len(self._data_results)))
            result = {'test': [], 'message':[], 'model': [], 'type': []}
            for k,v in self._test_results.items():    
                for a,b in v.items():
                    result['type'].append('test')
                    result['model'].append(k)
                    result['test'].append(a)
                    result['message'].append(b)
            for k,v in self._model_results.items():    
                for a,b in v.items():
                    result['type'].append('model')
                    result['model'].append(k)
                    result['test'].append(a)
                    result['message'].append(b)
            for k,v in self._data_results.items():    
                for a,b in v.items():
                    result['type'].append('data')
                    result['model'].append(k)
                    result['test'].append(a)
                    result['message'].append(b)
           
            display(pd.DataFrame.from_dict(result))
    def get_widget(self):
        return self._widget_main


import pailab.analysis.plot_helper as plt_helper
import pailab.analysis.plot as plot

# import plotly.graph_objs as go
# from plotly.offline import download_plotlyjs, init_notebook_mode, iplot


class Plotter:
    def _get_measures(self):
        measures = self._ml_repo.get_names(MLObjectType.MEASURE)
        result = []
        for m in measures:
            if self._model_selection.value in m:
                result.append(m)
        self._data_selection.options = result
        

    def _get_parameter(self, dummy=None):
        obj_cat = MLObjectType.TRAINING_PARAM
        if self._train_model_sel.value=='model':
            obj_cat = MLObjectType.MODEL_PARAM
        parameters = self._ml_repo.get_names(obj_cat)
        parameter = None
        for t in parameters:
            if self._model_selection.value in t:
                parameter = t
                break
        
        if parameter is None:
            self._param_sel.options = []
            return
        param = self._ml_repo.get(parameter)
        self._param_sel.options = [x for x in param.get_params().keys()]


    def __init__(self, ml_repo):
        self._ml_repo = ml_repo
        self._output = widgets.Output(layout=Layout(width='100%', height='100%'))
        self._model_selection = widgets.RadioButtons(
            options=self._ml_repo.get_names(MLObjectType.MODEL),
        #     value='pineapple',
            description='model:',
            disabled=False
        )

        self._train_model_sel = widgets.RadioButtons(
            options=['training', 'model'],
            value='model',
            description='param type:',
            disabled=False
        )
        self._param_sel = widgets.Select(
            description='parameter',
            disabled=False
        )
        self._get_parameter()

        self._train_model_sel.observe(self._get_parameter, 'value')
        self._data_selection = widgets.SelectMultiple(options=[])
        self._get_measures()
        
        self._plot_button = widgets.Button(description='plot')
        self._plot_button.on_click(self.plot)
        self._widget_main = widgets.VBox(  children=[ 
                widgets.HBox(
                    children=[self._plot_button,self._data_selection, self._model_selection, self._train_model_sel, self._param_sel]
                        ), 
                self._output] 
            )
        

    def get_widget(self):
        return self._widget_main

    def plot(self, d):
        with self._output:
            clear_output(wait=True)
            data = [ x for x in self._data_selection.value]
            if len(data) > 0 and self._param_sel.value is not None:
                use_train_param = True
                if self._train_model_sel.value == 'model':
                    use_train_param = False
                plot.measure_by_parameter(self._ml_repo, data, self._param_sel.value, data_versions= LAST_VERSION, training_param = use_train_param)




import os
import sqlite3
import pickle
from datetime import datetime, timedelta
import json
import pailab.ml_repo.repo_objects as repo_objects
from pailab.ml_repo.repo_store import RepoInfoKey, _time_from_version
import pailab.ml_repo.repo as repo
from pailab.ml_repo.disk_handler import RepoObjectDiskStorage
from git import Repo
import logging
logger = logging.getLogger(__name__)


class RepoObjectGitStorage(RepoObjectDiskStorage):
    """Object storage with git support.

    This storge stores all objects on disk  in a local git storage. It provides functionality to push and pull from 
    another git repo. Note that it handles all files in the same way as RepoObjectDiskStorage does (using methods from this storage).
    """

    @staticmethod
    def _is_git_repo(path):
        try:
            _ = Repo(path).git_dir
            return True
        except:
            return False

    def __init__(self, remote=None, **kwargs):
        """ Constructor
        Args:
            remote (str): The remote git repository. Defaults to None which means that there is no remote. If given and the target directory is not under git control, the repo will try to clon from the remote.
            folder (str): directory used to store the objects in files as well as the sqlite database
            file_format (str: 'pickle'|'json'): The fileformat used to save the objects. Defaults to 'pickle'.

        """

        super(RepoObjectGitStorage, self).__init__(**kwargs)
        # initialize git repo if it does not exist
        if not RepoObjectGitStorage._is_git_repo(self._main_dir):
            if remote is None:
                logger.info('Initialize new git repo.')
                _ = Repo.init(self._main_dir)
            else:
                # remove sqlite db to clone into directory
                if self._conn is not None:
                    self._conn.close()
                os.remove(self._sqlite_db_name())
                _ = Repo.clone_from(remote, self._main_dir)
                # check if cloned repository  contains sqlite db and if not create new one
                if not os.path.exists(self._sqlite_db_name()):
                    self._conn.close()
                self._setup_new()

    def _add(self, obj):
        """ Adds an object to the git repository

        Args:
            obj (RepoObject): the repo object to add to git
        """

        super(RepoObjectGitStorage, self)._add(obj)
        message = 'adding ' + obj['repo_info']['name']
        if (obj['repo_info']['commit_message'] is not None) and (obj['repo_info']['commit_message'] != ""):
            message = obj['repo_info']['commit_message']
        self.commit(message)

    def _delete(self, name, version):
        """ Delete an object from the repo

        Args:
            name (str): the identifier of the object
            version (str): the version of the object to delete
        """

        super(RepoObjectGitStorage, self)._delete(name, version)
        self.commit('deleting ' + name + ', version ' + version)

    def replace(self, obj):
        """ Overwrite existing object without incrementing version

        Args:
            obj (RepoObject): the repo object to be overwritten
        """

        super(RepoObjectGitStorage, self).replace(obj)
        self.commit('Replace object ' + obj['repo_info']['name'] +
                    ', version ' + obj['repo_info']['version'] + '.')

    def commit(self, message, force=True):
        """ Commits the changes

        Args:
            message (str): Commit message
            force (bool): If False, objecs will only be commited if integrity check succeeded.

        Raises:
            Exception: raises an exception if the integrity check fails
        """
        if not force:
            check = self.check_integrity()
            if len(check) > 0:
                raise Exception(
                    "Integrity check fails, cannot commit: " + str(check))
        git_repo = Repo(self._main_dir)
        git_repo.git.add('-A')
        git_repo.git.commit('-m', message)

    def push(self, remote_name='origin'):
        """ pushes the changes to the remote git repository

        Args:
            remote_name (str): name of the remote repository. Defaults to 'origin'.

        Raises:
            Exception: raises an exception if the remote does not exist
        """

        _git_repo = Repo(self._main_dir)
        remote = None
        for r in _git_repo.remotes:
            if r.name == remote_name:
                remote = r
                break
        if remote is None:
            raise Exception('Remote ' + remote_name + ' does not exist.')
        remote.push()

    def _merge_from_db(self, sqlite_db_2):
        """ merges the changes from the sqlite db

        Args:
            sqlite_db_2 (str): filename of the database to be added
        """

        c = self._conn.cursor()
        c.execute('ATTACH DATABASE "' + sqlite_db_2 + '" AS db_2')
        statement = 'INSERT OR IGNORE INTO versions(name, version, file, uuid_time) SELECT name, version, file, uuid_time FROM db_2.versions;'
        c.execute(statement)
        statement = 'INSERT OR IGNORE INTO mapping(name, category) SELECT name, category FROM db_2.mapping;'
        c.execute(statement)
        statement = 'INSERT OR IGNORE INTO modification_info(name, version, modifier, modifier_version) SELECT name, version, modifier, modifier_version FROM db_2.modification_info;'
        c.execute(statement)
        self._conn.commit()
        c.execute("DETACH DATABASE 'db_2';")

    def pull(self, remote_name='origin'):
        """ Pull from the remote git repository

        Args:
            remote_name (str): the name of the remote git repository. Defaults to 'origin'.

        Raises:
            Exception: raises an exception if the remote name is not available
            Exception: raises an error if the pull fails
        """

        # self._conn.close()
        remote = None
        _git_repo = Repo(self._main_dir)
        for r in _git_repo.remotes:
            if r.name == remote_name:
                remote = r
                break
        if remote is None:
            raise Exception('Remote ' + remote_name + ' does not exist.')
        self._conn.close()
        os.rename(self._sqlite_db_name(), self._sqlite_db_name() + '_old')
        try:
            remote.pull()
        except:
            os.rename(self._sqlite_db_name() + '_old', self._sqlite_db_name())
            raise Exception('An error occured during pull: ' + (str(e)))
        self._conn = sqlite3.connect(self._sqlite_db_name())
        self._merge_from_db(self._sqlite_db_name() + '_old')
        os.remove(self._sqlite_db_name() + '_old')



"""Module for pailab to pandas

    This module defines all necessary objects and functions to use pandas for preprocessing within pailab
    Currently only preprocessing functions without a previous fit to the training data is supported.
"""

__version__ = '0.0.1'

import pandas as pd

from pailab.ml_repo.repo_objects import repo_object_init
from pailab.ml_repo.repo_objects import RepoInfoKey
from pailab.ml_repo.repo_objects import Preprocessor
from pailab.ml_repo.repo import MLObjectType


def get_classname(x):
    return x.__class__.__module__ + '.' + x.__class__.__name__

# region Preprocessing


class PandasPreprocessingParam:
    """Interfaces the parameters of the pandas preprocessing algorithms

    """

    @repo_object_init()
    def __init__(self, preprocessor, pandas_param):
        self.pandas_module_name = preprocessor.__module__
        self.pandas_function_name = preprocessor.__name__
        self.pandas_params = pandas_param

    def get_params(self):
        if self.pandas_params is None:
            return {}
        else:
            return self.pandas_params


class PandasPreprocessor:
    """Class to store all pandas preprocessor
    """

    @repo_object_init()
    def __init__(self, preprocessor):
        self.preprocessor = preprocessor


def transform_pandas(preprocessor_param, data_x, column_names):
    def get_function(full_function_name):
        parts = full_function_name.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m
    m = get_function(preprocessor_param.pandas_module_name +
                     '.' + preprocessor_param.pandas_function_name)

    # tranform to pandas
    pandas_data = pd.DataFrame(data=data_x, columns=column_names)

    # transform
    pandas_data = m(data = pandas_data, **preprocessor_param.get_params())

    # transform back
    x_data = pandas_data.values
    x_coordinates = list(pandas_data)

    return x_data, x_coordinates


def add_preprocessor(repo, pandas_preprocessor, preprocessor_name=None, preprocessor_param=None):
    """Adds a new pandas preprocessor to a pailab MLRepo

    Args:
        repo ([type]): [description]
        preprocessor (Class): The pandas preprocessor class
        preprocessor_name ([type], optional): Defaults to None. [description]
        preprocessor_param ([type], optional): Defaults to None. [description]
    """

    # preprocessor name
    if not callable(pandas_preprocessor):
        raise Exception(
            'Currently only function are supported as pandas preprocessors, type was ' + type(pandas_preprocessor))

    p_name = preprocessor_name
    if p_name is None:
        p_name = pandas_preprocessor.__name__

    # add preprocessing
    repo.add_preprocessing_transforming_function(
        transform_pandas, repo_name=p_name + '/transform_pandas')

    pandas_param = PandasPreprocessingParam(pandas_preprocessor, preprocessor_param,
                                            repo_info={RepoInfoKey.NAME.value: p_name + '/preprocessor_param',
                                                       RepoInfoKey.CATEGORY: MLObjectType.PREPROCESSOR_PARAM.value})

    if pandas_param is not None:
        repo.add(pandas_param, 'adding preprocessor parameter')

    repo.add_preprocessor(p_name, transforming_function=p_name+'/transform_pandas', fitting_function=None,
                          preprocessor_param=p_name + '/preprocessor_param')

# endregion



# -*- coding: utf-8 -*-
"""This module contains all functions and classes for the MLTree. The MLTree buils a tree-like
structure of the objects in a given repository. This allows the user to access objects in a
comfortable way allowing for autocompletion (i.e. in Jupyter notebooks).

To use it one can simply call the :py:meth:`pailab.tools.tree.MLTree.add_tree` method to 
add such a tree to the current repository::

    >>from pailab.tools.tree import MLTree
    >>MLTree.add_tree(ml_repo)

After the tree has been added, one can simply use the tree. Here, using autocompletion makes the basic work wih repo objects quite simply.
Each tree node provides useful functions that can be applied:

- ``load`` loads the object of the given tree node or the child tree nodes of the current node. a
  After calling load the respective nodes have a new attribute ``obj`` that contains the respective loaded object. To load all objects belonging to the models subtree like 
  parameters, evaluations or measures one can call::

   >> ml_repo.tree.models.load()

- ``history`` lists the history of all objects of the respective subtree, where history excepts certain parameters such as a range of versions or 
  which repo object information to include. To list th history of all training data just use::

    >> ml_repo.tree.training_data.history()

- ``modifications`` lists all objects of the respective subtree that have been modified and no yet been committed.

There are also node dependent function (depending on what object the node represents).
"""
import logging
from numpy import load
from deepdiff import DeepDiff
from pailab.ml_repo.repo import MLObjectType, MLRepo
from pailab.ml_repo.repo_objects import RepoInfoKey, DataSet  # pylint: disable=E0401
from pailab.ml_repo.repo_store import RepoStore  # pylint: disable=E0401
import pailab.ml_repo.repo_store as repo_store
import pailab.ml_repo.repo_objects as repo_objects
logger = logging.getLogger(__name__)

#region collections and items



class _RepoObjectItem:

    def __init__(self, name, ml_repo, repo_obj = None):
        self._name = name
        self._repo = ml_repo
        if repo_obj is not None:
            self.obj = repo_obj
     
    def _set(self, path, items):
        if len(path) > 0:
            if len(path) == 1:
                setattr(self, path[0], items[0])
                return
            if hasattr(self, path[0]):
                getattr(self, path[0])._set(path[1:], items[1:])
            else:
                setattr(self, path[0], items[0])
                items[0]._set(path[1:], items[1:])

    def load(self, version=repo_store.LAST_VERSION, full_object=False,
            modifier_versions=None, containing_str=None):
            if containing_str is None or containing_str in self._name:
                if self._repo is not None:
                    self.obj = self._repo.get(self._name, version, full_object, modifier_versions, throw_error_not_exist = False)
            for v in self.__dict__.values():
                if hasattr(v,'load'):
                    v.load(version, full_object, modifier_versions, containing_str)

    def modifications(self, commit=False, commit_message=''):
        result = {}
        if self._name is not None:
            try:
                if self._repo is not None:
                    obj_orig = self._repo.get(
                        self.obj.repo_info[RepoInfoKey.NAME], version=self.obj.repo_info[RepoInfoKey.VERSION])
                    diff = DeepDiff(obj_orig, self.obj,
                                ignore_order=True)
            except AttributeError:
                return None
            if len(diff) == 0:
                return None
            else:
                if commit and (self._repo is not None):
                    version = self._repo.add(
                        self.obj, message=commit_message)
                    self.obj = self._repo.get(self._name, version=version)
                result = {self._name: diff}
        for v in self.__dict__.values():
            if hasattr(v, 'modifications'):
                tmp = v.modifications(commit, commit_message)
                if tmp is not None:
                    result.update(tmp)
        return result

    def history(self, version = (repo_store.FIRST_VERSION,repo_store.LAST_VERSION), 
                repo_info = [RepoInfoKey.NAME, RepoInfoKey.AUTHOR, RepoInfoKey.COMMIT_DATE, RepoInfoKey.COMMIT_MESSAGE], 
                obj_data = []):
        history = []
        if self._repo is not None:
            history = self._repo.get(self._name, version = version,  throw_error_not_exist=False)
        if not isinstance(history, list):
            history = [history]
        result = {}
        tmp = []
        for h in history:
            r = {}
            for r_info in repo_info:
                r[str(r_info)] = h.repo_info[r_info]
            for o_info in obj_data:
                r[o_info] = obj_data.__dict__[o_info]
            tmp.append(r)
        result[self._name] = tmp
        for v in self.__dict__.values():
            if isinstance(v, _RepoObjectItem):
                tmp2 = v.history(version, repo_info, obj_data)
                if tmp2 is not None:
                    result.update(tmp2)
        if len(result) > 0:
            return result
        

    def __call__(self, containing_str=None):
        # if len(self.__dict__) == 1:
        if containing_str is not None:
            result = []
            if containing_str in self._name:
                result.append(self._name)
            for v in self.__dict__.values():
                if isinstance(v, _RepoObjectItem):
                    d = v(containing_str)
                    if isinstance(d, str):
                        result.append(d)
                    else:
                        result.extend(d)
            return [x for x in result if containing_str in x]
        else:
            return self._name
        return result


class _RawDataItem(_RepoObjectItem):
    def __init__(self, name, ml_repo, repo_obj = None):
        super(_RawDataItem,self).__init__(name, ml_repo, repo_obj)

    def append(self, x_data, y_data = None):
        """Append data to a RawData object

        It appends data to the given RawData object and updates all training and test DataSets which implicitely changed by this update.

        Args:
            name (string): name of RawData object
            x_data (numpy matrix): the x_data to append
            y_data (numpy matrix, optional): Defaults to None. The y_data to append
        
        Raises:
            Exception: If the data is not consistent to the RawData (e.g. different number of x-coordinates) it throws an exception.
        """
        logger.info('Start appending ' + str(x_data.shape[0]) + ' datapoints to RawData' + self._name)
        raw_data = self._repo.get(self._name)
        if len(raw_data.x_coord_names) != x_data.shape[1]:
            raise Exception('Number of columns of x_data of RawData object is not equal to number of columns of additional x_data.')
        if raw_data.y_coord_names is None and y_data is not None:
            raise Exception('RawData object does not contain y_data but y_data is given')
        if raw_data.y_coord_names is not None:
            if y_data is None:
                raise Exception('RawData object has y_data but no y_data is given')
            if y_data.shape[1] != len(raw_data.y_coord_names ):
                raise Exception('Number of columns of y_data of RawData object is not equal to number of columns of additional y_data.')
        numpy_dict = {'x_data' : x_data}
        if raw_data.y_coord_names is not None:
            numpy_dict['y_data'] =  y_data
        raw_data.n_data += x_data.shape[0]
        old_version = raw_data.repo_info[RepoInfoKey.VERSION]
        new_version = self._repo.add(raw_data)
        self._repo._numpy_repo.append(self._name, old_version, new_version, numpy_dict)
        # now find all datasets which are affected by the updated data
        changed_data_sets = []
        training_data = self._repo.get_training_data(full_object = False)
        if isinstance(training_data, DataSet):
            if training_data.raw_data == self._name and training_data.raw_data_version == repo_store.RepoStore.LAST_VERSION:
                if training_data.end_index is None or training_data.end_index < 0:
                    training_data.raw_data_version = new_version
                    changed_data_sets.append(training_data)
        test_data = self._repo.get_names(MLObjectType.TEST_DATA)
        for d in test_data:
            data = self._repo.get(d)
            if isinstance(data, DataSet):
                if data.raw_data == self._name and data.raw_data_version == repo_store.RepoStore.LAST_VERSION:
                    if data.end_index is None or data.end_index < 0:
                        data.raw_data_version = new_version
                        changed_data_sets.append(data)
        self._repo.add(changed_data_sets, 'RawData ' + self._name + ' updated, add DataSets depending om the updated RawData.')
        if hasattr(self, 'obj'):#update current object
            self.obj = self._repo.get(self._name, version=new_version)
        logger.info('Finished appending data to RawData' + self._name)

class _RawDataCollection(_RepoObjectItem):
    @staticmethod
    def __get_name_from_path(path):
        return path.split('/')[-1]

    def __init__(self, repo):
        super(_RawDataCollection, self).__init__('raw_data', repo)
        names = repo.get_names(MLObjectType.RAW_DATA)
        for n in names:
            setattr(self, _RawDataCollection.__get_name_from_path(n), _RawDataItem(n, repo))
        
    def add(self, name, data, input_variables = None, target_variables = None):
        """Add raw data to the repository

        Arguments:
            data_name {name of data} -- the name of the data added
            data {pandas datatable} -- the data as pandas datatable
        
        Keyword Arguments:
            input_variables {list of strings} -- list of column names defining the input variables for the machine learning (default: {None}). If None, all variables are used as input
            target_variables {list of strings} -- list of column names defining the target variables for the machine learning (default: {None}). If None, no target data is added from the table.
        """
        path = 'raw_data/' + name

        if input_variables is None:
            input_variables = list(data)
            if not target_variables is None:
                [input_variables.remove(x) for x in target_variables]
        else:
            # check whether the input_variables are included in the data
            if not [item for item in input_variables if item in list(data)] == input_variables:
                raise Exception('RawData does not include at least one column included in input_variables')
      
        if target_variables is not None:
            # check if target variables are in list
            if not [item for item in target_variables if item in list(data)] == target_variables:
                raise Exception('RawData does not include at least one column included in target_variables')
            raw_data = repo_objects.RawData(data.loc[:, input_variables].values, input_variables, data.loc[:, target_variables].values, 
                target_variables, repo_info = {RepoInfoKey.NAME: path})
        else:
            raw_data = repo_objects.RawData(data.loc[:, input_variables].values, input_variables, repo_info = {RepoInfoKey.NAME: path})
        v = self._repo.add(raw_data, 'data ' + path + ' added to repository' , category = MLObjectType.RAW_DATA)
        obj = self._repo.get(path, version=v, full_object = False)
        setattr(self, name, _RawDataItem(path, self._repo, obj))

    def add_from_numpy_file(self, name, filename_X, x_names, filename_Y=None, y_names = None):
        path = name
        X = load(filename_X)
        Y = None
        if filename_Y is not None:
            Y = load(filename_Y)
        raw_data =  repo_objects.RawData(X, x_names, Y, y_names, repo_info = {RepoInfoKey.NAME: path})
        v = self._repo.add(raw_data, 'data ' + path + ' added to repository' , category = MLObjectType.RAW_DATA)
        obj = self._repo.get(path, version=v, full_object = False)
        setattr(self, name, _RawDataItem(path, self._repo, obj))

class _TrainingDataCollection(_RepoObjectItem):

    @staticmethod
    def __get_name_from_path(path):
        return path.split('/')[-1]
    
    def __init__(self, repo):
        super(_TrainingDataCollection, self).__init__('training_data', None)
        self.__repo = repo # we store ml_repo in __repo to circumvent that obj is loaded from eneric base class
        names = repo.get_names(MLObjectType.TRAINING_DATA)
        for n in names:
            setattr(self, _TrainingDataCollection.__get_name_from_path(n), _RepoObjectItem(n, repo))
        
    def add(self, name, raw_data, start_index=0, 
        end_index=None, raw_data_version='last'):
        #path = 'training_data/' + name
        data_set = repo_objects.DataSet(raw_data, start_index, end_index, 
                raw_data_version, repo_info = {RepoInfoKey.NAME: name, RepoInfoKey.CATEGORY: MLObjectType.TRAINING_DATA})
        v = self.__repo.add(data_set)
        tmp = self.__repo.get(name, version=v)
        item = _RepoObjectItem(name, self.__repo, tmp)
        setattr(self, name, item)

class _TestDataCollection(_RepoObjectItem):
    @staticmethod
    def __get_name_from_path(path):
        return path.split('/')[-1]
    
    def __init__(self, repo):
        super(_TestDataCollection, self).__init__('test_data', None)
        self.__repo = repo # we store ml_repo in __repo to circumvent that obj is loaded from eneric base class
        names = repo.get_names(MLObjectType.TEST_DATA)
        for n in names:
            setattr(self, _TestDataCollection.__get_name_from_path(n), _RepoObjectItem(n,repo))
        
    def add(self, name, raw_data, start_index=0, 
        end_index=None, raw_data_version='last'):
        data_set = repo_objects.DataSet(raw_data, start_index, end_index, 
                raw_data_version, repo_info = {RepoInfoKey.NAME: name, RepoInfoKey.CATEGORY: MLObjectType.TEST_DATA})
        v = self.__repo.add(data_set)
        tmp = self.__repo.get(name, version=v)
        item = _RepoObjectItem(name, self.__repo, tmp)
        setattr(self, name, item)

class _MeasureItem(_RepoObjectItem):
    def __init__(self, name, ml_repo, repo_obj = None):
        super(_MeasureItem, self).__init__(name, ml_repo, repo_obj) 

class _JobItem(_RepoObjectItem):
    def __init__(self, name, ml_repo, repo_obj = None):
        super(_JobItem, self).__init__(name, ml_repo, repo_obj) 

class _MeasureCollection(_RepoObjectItem):
    def __init__(self, name, ml_repo):
        super(_MeasureCollection, self).__init__('measures', None)
        names = ml_repo.get_names(MLObjectType.MEASURE)
        for n in names:
            path = n.split('/')[2:]
            items = [None] * len(path)
            for i in range(len(items)-1):
                items[i] = _RepoObjectItem(path[i], None)
            items[-1] = _MeasureItem(n, ml_repo)
            self._set(path, items)
            #items[-2] = MeasuresOnDataItem

class _EvalCollection(_RepoObjectItem):
    def __init__(self, name, ml_repo):
        super(_EvalCollection, self).__init__('eval', None)
        names = ml_repo.get_names(MLObjectType.EVAL_DATA)
        for n in names:
            path = n.split('/')[2:]
            items = [None] * len(path)
            for i in range(len(items)-1):
                items[i] = _RepoObjectItem(path[i], None)
            items[-1] = _MeasureItem(n, ml_repo)
            self._set(path, items)

class _TestCollection(_RepoObjectItem):
    def __init__(self, name, ml_repo):
        super(_TestCollection, self).__init__('tests', None)
        names = ml_repo.get_names(MLObjectType.TEST)
        for n in names:
            path = n.split('/')[2:]
            items = [None] * len(path)
            for i in range(len(items)-1):
                items[i] = _RepoObjectItem(path[i], None)
            items[-1] = _RepoObjectItem(n, ml_repo)
            self._set(path, items)

class _JobCollection(_RepoObjectItem):
    def __init__(self, name, ml_repo, model_name):
        super(_JobCollection, self).__init__('jobs', None)
        names = ml_repo.get_names(MLObjectType.JOB)
        for n in names:
            if model_name in n:
                path = n.split('/')
                path = path[path.index('jobs')+1:]
                items = [None] * len(path)
                for i in range(len(items)-1):
                    items[i] = _RepoObjectItem(path[i], None)
                items[-1] = _JobItem(n, ml_repo)
                self._set(path, items)

class _ModelItem(_RepoObjectItem):
    def __init__(self, name, ml_repo, repo_obj = None):
        super(_ModelItem,self).__init__(name, ml_repo, repo_obj)
        self.model = _RepoObjectItem(name + '/model', ml_repo)
        self.eval = _EvalCollection(name + '/eval', ml_repo)
        self.model_param = _RepoObjectItem(name + '/model_param', ml_repo)
        self.tests = _TestCollection(name + '/tests', ml_repo)
        self.measures = _MeasureCollection(name+ '/measure', ml_repo)
        self.jobs = _JobCollection(name+'/jobs', ml_repo, name)
        if ml_repo._object_exists(name+'/training_stat'):
            self.training_statistic = _RepoObjectItem(name+'/training_stat', ml_repo)
        if ml_repo._object_exists(name+'/training_param'):
            self.training_param = _RepoObjectItem(name + '/training_param', ml_repo)


    def set_label(self, label_name, version = repo_store.RepoStore.LAST_VERSION, message=''):
        self._repo.set_label(label_name, self._name+ '/model', version, message)

class _LabelCollection(_RepoObjectItem):
    def __init__(self, repo):
        super(_LabelCollection,self).__init__('labels', None)
        names = repo.get_names(MLObjectType.LABEL)
        for n in names:
            #label = ml_repo.get()
            setattr(self, n, _RepoObjectItem(n, repo))
        
class _ModelCollection(_RepoObjectItem):
    @staticmethod
    def __get_name_from_path(name):
        return name

    def __init__(self, repo):
        super(_ModelCollection,self).__init__('models', None)
        names = repo.get_names(MLObjectType.MODEL)
        for n in names:
            setattr(self, _ModelCollection.__get_name_from_path(n), _ModelItem(n, repo))
        self.labels = _LabelCollection(repo)
        
    def add(self, name):
        setattr(self, name, _ModelItem(name,self._repo))



class _CacheDataCollection(_RepoObjectItem):

    @staticmethod
    def __get_name_from_path(path):
        return path.split('/')[-1]
    
    def __init__(self, repo):
        super(_CacheDataCollection, self).__init__('cache', None)
        self.__repo = repo # we store ml_repo in __repo to circumvent that obj is loaded from eneric base class
        names = repo.get_names(MLObjectType.CACHED_VALUE)
        for n in names:
            setattr(self, _CacheDataCollection.__get_name_from_path(n), _RepoObjectItem(n, repo))
#endregion



class MLTree:

    @staticmethod
    def add_tree(ml_repo):
        """Adds an MLTree to a repository.

        Args:
            ml_repo (MLRepo): the repository the tre is added
        """
        setattr(ml_repo, 'tree', MLTree(ml_repo))
        ml_repo._add_triggers.append(ml_repo.tree.reload)
        
    def __create(self):
        self.raw_data = _RawDataCollection(self.__ml_repo)
        self.training_data = _TrainingDataCollection(self.__ml_repo)
        self.test_data = _TestDataCollection(self.__ml_repo)
        self.models = _ModelCollection(self.__ml_repo)
        self.cache = _CacheDataCollection(self.__ml_repo)

    def __init__(self, ml_repo):
        self.__ml_repo = ml_repo
        self.__create()

    def reload(self, **kwargs):
        """Method to reload the tree after objects have been added or deleted from the repository.
        """
        self.__create()  # todo make this more efficient by just updating collections and items which are affected by this

    def modifications(self):
        """Return a dictionary of all objects that were modified but no yet 
        commited to the repository.
        
        Returns:
            dict: dictionary mapping object ids to dictionary of the modified attributes 
        """
        result = {}
        tmp = self.raw_data.modifications()
        if tmp is not None:
            result.update(tmp)
        tmp = self.training_data.modifications()
        if tmp is not None:
            result.update(tmp)
        tmp = self.test_data.modifications()
        if tmp is not None:
            result.update(stmp)
        tmp = self.models.modifications()
        if tmp is not None:
            result.update(tmp)
        if len(result) == 0:
            return None
        return result

        



"""This module provides out-of-the box plots to analyse models whee titles, axes labels, additional information is automatically
added to the resuling figures from the information stored in the repository.
"""

import logging
import math
import numpy as np
import pandas as pd
import pailab.analysis.plot_helper as plot_helper  # pylint: disable=E0611
from pailab.ml_repo.repo_store import RepoStore, LAST_VERSION
from pailab.ml_repo.repo import NamingConventions, MLObjectType  # pylint: disable=E0611,E0401
from pailab import RepoInfoKey

has_plotly = True
try:
    from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot
    from plotly import tools
    import plotly.graph_objs as go
except ImportError:
    has_plotly = False

logger = logging.getLogger(__name__)

init_notebook_mode(connected=True)


def measure_by_parameter(ml_repo, measure_name, param_name, data_versions=None, training_param=False, logscale_y=False, logscale_x=False):
    """Plot a measure value vs a certain training or model parameter.

    Args:
        ml_repo (MLRepo): MLRepo
        measure_name (str): Name of measure to be plotted.
        param_name (str): Name of parameter to be plotted. To define a subparameter on can use the '/' to define the path to the parameter.
        data_versions (str, optional): Version of the dataset that should be underlying the measure. If Noe, the latest version for the underlying data is used.
                                        Defaults to None.
        training_param (bool, optional): Boolean that defines if parameter of interest belongs to training or model parameter. Defaults to False.
        logscale_y (bool): If true, the y-axis will be log scale. Defaults to False.
        logscale_x (bool): If true, the x-axis will be log scale. Defaults to False.

    Examples:
        To plot the maximum error (which must have been defined in the measures) for the model ``DecisionTreeRegressor`` on the dataset ``sample1`` against
        the parameter ``learning_rate`` contained in the subparameter ``optim_param`` we may call::

            >> measure_by_parameter(ml_repo, 'DecisionTreeRegressor/measure/sample1/max', 'optim_param/learning_rate')
    """
    if logscale_x:
        x_scaler = math.log10
    else:
        x_scaler = lambda x: x

    if logscale_y:
        y_scaler = math.log10
    else:
        y_scaler = lambda x: x

    x = plot_helper.get_measure_by_parameter(
        ml_repo, measure_name, param_name, data_versions, training_param)
    data = []
    model_label_annotations = []
    for k, measures in x.items():
        data_name = str(NamingConventions.Data(
            NamingConventions.EvalData(NamingConventions.Measure(measure_name))))
        data_versions = set()

        for measure in measures:
            data_versions.add(measure['data_version'])
            if 'model_label' in measure:
                model_label_annotations.append(dict(x=x_scaler(measure[param_name]), y=y_scaler(measure['value']), xref='x', yref='y', text=measure['model_label'],
                                                    showarrow=True,
                                                    arrowhead=2,
                                                    # ax=0,
                                                    # ay=-30
                                                    ))
        measures = pd.DataFrame(measures)
        for d_version in data_versions:
            # if True:
            df = measures.loc[measures['data_version'] == d_version]
            text = ["model version: " + str(x['model_version']) + '<br>' +
                    data_name + ': ' + str(x['data_version']) + '<br>'
                    + 'train_data: ' + str(x['train_data_version'])
                    for index, x in df.iterrows()]

            if True:  # len(x) > 1:
                plot_name = k + ': ' + str(d_version)
            # else:
            #    plot_name = data_name + ': ' + str(d_version)
            data.append(
                go.Scatter(
                    x=df[param_name],
                    y=df['value'],
                    text=text,
                    name=plot_name,
                    mode='markers'
                )

            )

    xaxis = dict(title=param_name)
    if logscale_x:
        xaxis['type'] = 'log'
    yaxis = dict(title=NamingConventions.Measure(
            measure_name).values['measure_type'])
    if logscale_y:
        yaxis['type'] = 'log'
    layout = go.Layout(
        title='measure by parameter',
        annotations=model_label_annotations,
        xaxis=xaxis,
        yaxis=yaxis
    )
    # IPython notebook
    # py.iplot(data, filename='pandas/basic-line-plot')

    fig = go.Figure(data=data, layout=layout)
    # return fig
    iplot(fig)  # , filename='pandas/basic-line-plot')


def projection(ml_repo, left, right, n_steps=100, model=None, labels=None,  output_index=None, direction=None):
    logger.info('Start projection with ' + str(n_steps) + ' steps.')
    x = plot_helper.project(ml_repo, model, labels, left,
                            right, output_index=output_index, n_steps=n_steps)
    # use training data to get output name
    training = ml_repo.get_names(MLObjectType.TRAINING_DATA)
    output_name = ml_repo.get(training[0]).y_coord_names[0]
    data = []
    x_data = [0.0 + float(i)/float(n_steps-1) for i in range(n_steps)]
    for k, v in x.items():
        data.append(
                    go.Scatter(
                        x=x_data,
                        y=v,
                        name=k
                    )
        )
    layout = go.Layout(
        title='projection',
        xaxis=dict(title='steps'),
        yaxis=dict(title=output_name)
    )
    fig = go.Figure(data=data, layout=layout)
    iplot(fig)


def measure_history(ml_repo, measure_name, logscale_y=False):
    """Plots the history of the model w.r.t. a defined measure. The x-axis is defined by the indert datetime of each model.


    Args:
        ml_repo (MLRepo): MLRepo.
        measure_name (str, iterable of str): Name (or iterable of names) of measure(s) to plot (a measure name includes the name of the underlying model and dataset).
        logscale_y (bool): If true, the y-axis will be log scale. Defaults to False.

    Examples:
        To plot the history of the maximum error (which must have been defined in the measures) for the model ``DecisionTreeRegressor`` on the dataset ``sample1``::

            >> measure_history(ml_repo, 'DecisionTreeRegressor/measure/sample1/max')
    """
    if logscale_y:
        y_scaler = math.log10
    else:
        y_scaler = lambda x: x

    x = plot_helper.get_measure_history(
        ml_repo, measure_name)
    data = []
    model_label_annotations = []
    for k, measures in x.items():
        data_name = str(NamingConventions.Data(
            NamingConventions.EvalData(NamingConventions.Measure(measure_name))))
        data_versions = set()

        for measure in measures:
            data_versions.add(measure['data_version'])
            if 'model_label' in measure:
                model_label_annotations.append(dict(x=str(measure['datetime']), y=y_scaler(measure['value']), xref='x', yref='y', text=measure['model_label'],
                                                    showarrow=True,
                                                    arrowhead=2,  # 1
                                                    # ax=,
                                                    # ay=-30
                                                    ))
        measures = pd.DataFrame(measures)
        for d_version in data_versions:
            # if True:
            df = measures.loc[measures['data_version'] == d_version]
            text = ["model version: " + str(x['model_version']) + '<br>' +
                    data_name + ': ' + str(x['data_version']) + '<br>'
                    + 'train_data: ' + str(x['train_data_version'])
                    for index, x in df.iterrows()]

            if True:  # len(x) > 1:
                plot_name = k + ': ' + str(d_version)
            # else:
            #    plot_name = data_name + ': ' + str(d_version)
            data.append(
                go.Scatter(
                    x=df['datetime'],
                    y=df['value'],
                    text=text,
                    name=plot_name,
                    mode='markers'
                )
            )

    yaxis = dict(title=NamingConventions.Measure(
            measure_name).values['measure_type'])
    if logscale_y:
        yaxis['type'] = 'log'
    layout = go.Layout(
        title='measure history',
        annotations=model_label_annotations,
        xaxis=dict(title='t'),
        yaxis=yaxis
    )
    # IPython notebook
    # py.iplot(data, filename='pandas/basic-line-plot')
    fig = go.Figure(data=data, layout=layout)

    iplot(fig)  # , filename='pandas/basic-line-plot')


def _histogram(plot_dict, n_bins=None, histnorm='percent'):
    layout = go.Layout(
        title=plot_dict['title'],
        xaxis=dict(title=plot_dict['x0_name']),
        barmode='overlay'
    )
    plot_data = []
    opacity = 1.0
    if len(plot_dict['data'].keys()) > 1:
        opacity = 0.5
    for k, x in plot_dict['data'].items():
        text = ''
        for l, w in x['info'].items():
            text += l + ':' + str(w) + '<br>'
        if 'label' in x.keys():
            k = x['label'] + ', ' + k
        if n_bins is None:
            plot_data.append(go.Histogram(x=x['x0'],
                                        text=text,
                                        name=k,
                                        opacity=opacity,
                                        histnorm=histnorm))
        else:
            plot_data.append(go.Histogram(x=x['x0'],
                                        text=text,
                                        name=k,
                                        opacity=opacity,
                                        nbinsx=n_bins,
                                        histnorm=histnorm))

    fig = go.Figure(data=plot_data, layout=layout)

    iplot(fig)  # , filename='pandas/basic-line-plot')


def histogram_model_error(ml_repo, models, data_name, y_coordinate=None, data_version=LAST_VERSION, n_bins=None,  start_index=0, end_index=-1):
    """Plot histogram of differences between predicted and real values.

    The method plots histograms between predicted and real values of a certain target variable for reference data and models.
    The reference data is described by the data name and the version of the data (as well as the targt variables name). The models can be described
    by
    - a dictionary of model names to versions (a single version number, a range of versions or a list of versions)
    - just a model name (in this case the latest version is used)

    Args:
        ml_repo (MLRepo): [description]
        models (str or dict): A dictionary of model names to versions (a single version number, a range of versions or a list of versions) or
                just a model name (in this case the latest version is used)

        data_name (str or list of str): Name of input data to be used for the error plot.
        y_coordinate (int, str or list, optional): Index or (list of) name(s) of y-coordinate(s) used for error measurement. If None, the first coordinate is used. Defaults to None.
        data_version (str, optional): Version of the input data used. Defaults to LAST_VERSION.

    Examples:
        Plot histograms for errors in the variable mickey_mouse on the dataset my_test_data for the latest version of model_1 and all versions of model_2.

        >>> histogram_model_error(repo, models = {'model_1': ['latest'], 'model_2': ('first','latest')},
            data_name = 'my_test_data', y_coordinate='mickey_mouse')

        Plot histogram for error of latest version of model_2 on the latest version of my_test_data. Note that the plot would be empty if the latest version of model_2
        has not yet been evaluated on the latest version of my_test_data.

        >>> histogram_model_error(repo, models = 'model_2', data_name = 'my_test_data', y_coordinate='mickey_mouse')


    """
    if isinstance(y_coordinate, list):
        plot_dict = {'data':{}, 'x0_name': 'pointwise error'}
        for coord in y_coordinate:
            tmp = plot_helper.get_pointwise_model_errors(
                ml_repo, models, data_name, coord, start_index=start_index, end_index=end_index)
            for k,v in tmp['data'].items():
                plot_dict['data'][str(coord) + ':' + k] = v
        plot_dict['title'] = 'pointwise errors'
    else:    
        plot_dict = plot_helper.get_pointwise_model_errors(
            ml_repo, models, data_name, y_coordinate, start_index=start_index, end_index=end_index)
    _histogram(plot_dict, n_bins)


def scatter_model_error(ml_repo, models, data_name, x_coordinate, y_coordinate=None, start_index=0, end_index=-1):
    '''Plots for each model the pointwise error along a specified target dimension w.r.t. a specified input dimension.

    Args:
        ml_repo (MLRepo): MLRepo.
        models (str, iterable of str or dict of string to string): Definition of the models for whih the pointwise error are plotted.
            It may be a string defining the model or label name (in case of a model the latest model is used), an iterable of strings where again each string is either a model name
            or a label, or a dict mapping model names to version numbers.
        data_name (str or iterable of str): String or iterable of string defining the datasets over which the pointwise errors will be computed.
        x_coordinate (str or int): Either string or int defining the x-coordinate to which the error will be plotted.
        y_coordinate (str or int, optional): Either string or int defining the x-coordinate to which the error will be plotted. Defaults to None (then the 0th cordinate will be used).
    '''

    plot_dict = plot_helper.get_pointwise_model_errors(
        ml_repo, models, data_name, y_coordinate, x_coord_name=x_coordinate, start_index=start_index, end_index=end_index)

    layout = go.Layout(
        title=plot_dict['title'],
        xaxis=dict(title=plot_dict['x0_name']),
        yaxis=dict(title=plot_dict['x1_name'])
    )
    plot_data = []
    for k, x in plot_dict['data'].items():
        text = ''
        for l, w in x['info'].items():
            text += l + ':' + str(w) + '<br>'
        if 'label' in x.keys():
            k = x['label'] + ', ' + k
        plot_data.append(go.Scatter(x=x['x0'],
                                    y=x['x1'],
                                    text=text,
                                    name=k,
                                    mode='markers'))

    # IPython notebook
    # py.iplot(data, filename='pandas/basic-line-plot')
    fig = go.Figure(data=plot_data, layout=layout)

    iplot(fig)  # , filename='pandas/basic-line-plot')


def histogram_data(ml_repo, data, x_coordinate, n_bins=None,  start_index=0, end_index=-1):
    '''Plot the histogram of the input data along a specified coordinate direction.

    Args:
        ml_repo (MLRepo): MLRepo.
        data (str or dict): Either a string with the name of the data to be plotted (latest data will be plotte) or a dictionary of data names to version or list of versions.
        x_coordinate (str): String defining the x_coordinate to be plotted. Defaults to None.
    '''
    plot_dict = plot_helper.get_data(
        ml_repo, data, x_coordinate, start_index=start_index, end_index=end_index)
    _histogram(plot_dict, n_bins=n_bins)


def histogram_data_conditional_error(ml_repo, models, data, x_coordinate, y_coordinate=0,
                                    start_index=0, end_index=-1, percentile=0.1, n_bins=None, n_hist=1,
                                    metric = 'rbf',  **kwds):
    """Plots the distribution of input data along a given axis for the largest absolute pointwise errors in comparison to the distribution of all data.
    
    Args:
        ml_repo (MLRepo): repository
        models (str or dict): A dictionary of model names (or labels) to versions (a single version number, a range of versions or a list of versions) or 
            just a model name (in this case the latest version is used)
        data (str): name of dataset used for plotting
        x_coordinate (str): Name of x coordinate for which the distribution will be plotted. If None, the method tries to find interesting coordinates where distribution of perentile differs to original distribution.
        y_coordinate (str, optional): Name of y-coordinate for which the error is determined. Defaults to 0 (use first y-coordinate). 
            If None, the method tries to find interesting coordinates where distribution of perentile differs to original distribution.
        start_index (int, optional): Defaults to 0. Startindex of data.
        end_index (int, optional): Defaults to -1. Endindex of data.
        percentile (float, optional): Defaults to 0.1. Percentage of largest absolute errors used.
        n_bins (int, optional): Defaults to None. Number of bin of histogram.
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array. defaults to 'rbf'.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        **kwds: optional keyword parameters that are passed directly to the kernel function.
    
    """
    if x_coordinate is None or y_coordinate is None:
        tmp =  pd.DataFrame.from_dict( plot_helper.get_ptws_error_dist_mmd(ml_repo, models, data, x_coordinate,
                     y_coordinate, start_index=start_index, end_index=end_index, percentile = percentile,  metric='rbf',  **kwds)
            )
        tmp = tmp.sort_values(['mmd'], ascending = False)
        recommended_coordinates = set()
        for i in range(tmp.shape[0]):
            recommended_coordinates.add((tmp.iloc[i]['x-coord'], tmp.iloc[i]['y-coord'], ))
            if len(recommended_coordinates) > n_hist:
                break
        for coord in recommended_coordinates:
            histogram_data_conditional_error(ml_repo, models, data, coord[0], coord[1],
                start_index, end_index, percentile, n_bins, n_hist)
    else:
        tmp = plot_helper.get_pointwise_model_errors(
            ml_repo, models, data, y_coordinate, x_coord_name=x_coordinate, start_index = start_index, end_index = end_index)
        
        plot_data = {}
            
        for k, x in tmp['data'].items():
            abs_err = np.abs(x['x1'])
            sorted_indices = np.argsort(abs_err)
            i_start = int( (1.0-percentile)*len(sorted_indices))
            indices = sorted_indices[i_start:]
            data = {'x0': x['x0'][indices], 'info': x['info']}
            if 'label' in x.keys():
                plot_data[x['label']+':'+str(percentile)] = data
                plot_data[x['label']] = {'x0': x['x0'][start_index:end_index], 'info': x['info']}
            else:
                plot_data[k+':'+str(percentile)] = data
                plot_data[k] = {'x0': x['x0'][start_index:end_index], 'info': x['info']}
        plot_dict = {'data': plot_data, 'title': tmp['title'] + ', inlcuding distribution w.r.t ' + str(100*percentile) + ' % quantile', 'x0_name' : tmp['x0_name'], 'x1_name': tmp['x1_name']}
        _histogram(plot_dict, n_bins=n_bins)


def _ice_plotly(ice_results, ice_points = None, height = None, width = None, ice_results_2 = None, clusters = None):
    data = []
    if ice_points is None:
        ice_points = range(ice_results.ice.shape[0])
    if clusters is not None:
        ice_points_tmp = []
        for i in ice_points:
            if ice_results.labels[i] in clusters:
                ice_points_tmp.append(i)
        _ice_plotly(ice_results, ice_points=ice_points_tmp, ice_results_2 = ice_results_2, height=height, width=width)
        return

    # plot ice curves
    for i in ice_points:
        data.append(go.Scatter(x=ice_results.x_values, y = ice_results.ice[i,:], name = ice_results.data_name + '[' + str(ice_results.start_index +i) + ',:]'))
        if ice_results_2 is not None:
            data.append(go.Scatter(x=ice_results_2.x_values, y = ice_results_2.ice[i,:], name = 'model2, ' + ice_results.data_name + '[' + str(ice_results.start_index +i) + ',:]'))
    layout = go.Layout(
        title='ICE, model: ' + ice_results.model + ', version: ' + ice_results.model_version,
        xaxis=dict(title=ice_results.x_coord_name),
        yaxis=dict(title=ice_results.y_coord_name),
        height = height,
        width = width
    )
    fig = go.Figure(data=data, layout=layout)
    iplot(fig)  # , filename='pandas/basic-line-plot')


def _ice_clusters_plotly(ice_results, height = None, width = None, ice_results_2= None, clusters = None):
    data = []
    if clusters is None:
        clusters = range(ice_results.cluster_centers.shape[0])
    # plot ice cluster curves
    for i in clusters:
        data.append(go.Scatter(x=ice_results.x_values, y = ice_results.cluster_centers[i,:], name = 'cluster ' + str(i)))
    if ice_results_2 is not None:
        cluster_averages = ice_results.compute_cluster_average(ice_results_2)
        for i in clusters:
            data.append(go.Scatter(x=ice_results.x_values, y = cluster_averages[i,:], name = 'average ' + str(i)))

    layout = go.Layout(
        title='ICE clusters, model: ' + ice_results.model + ', version: ' + ice_results.model_version,
        xaxis=dict(title=ice_results.x_coord_name),
        yaxis=dict(title=ice_results.y_coord_name),
        height = height,
        width = width
    )
    # fig = tools.make_subplots(rows=2, cols=2, subplot_titles=('Plot 1', 'Plot 2',
    #                                                      'Plot 3', 'Plot 4'))#, layout = layout)
    # fig['layout'].update(title='ICE, model: ' + ice_results.model + ', version: ' + ice_results.model_version, xaxis=dict(title=ice_results.x_coord_name),
    #    yaxis=dict(title=ice_results.y_coord_name),
    #    height = height,
    #    width = width)
    fig = go.Figure(data=data, layout=layout)
    # fig.append_trace(data,1,1)
    iplot(fig)  # , filename='pandas/basic-line-plot')


def ice(ice_results, height = None, width = None, ice_points = None, ice_results_2 = None, clusters = None):
    """Plots ICE graphs computed with tools.interpretation.compute_ice.
    
    Args:
        ice_results (tools.interpretation.ICE_Results): Resulting object after calling tools.interpretation.compute_ice.
        height ([type], optional): [description]. Defaults to None.
        width ([type], optional): [description]. Defaults to None.
        ice_points (list of int, optional): List of integers to select the ICE graphs to be plotted. Defaults to None (all ICE graphs will be plotted).
        ice_results_2 (tools.interpretation.ICE_Results, optional): ICE graphs for another model which are plotted for comparison. Defaults to None.
        clusters (list of int, optional): List of clusters of ICE graphs that will be plotted. In order to wok with this feature, 
                    the ice_result must have been derived calling  tools.interpretation.compute_ice so that clustering will be executed (cluster params need to be set).
                    Defaults to None (no clusters plotted).
    """
    if ice_results_2 is not None:
        ice_results._validate_for_comparison(ice_results_2)
    if has_plotly:
        _ice_plotly(ice_results, height=height, width=width, ice_points = ice_points, ice_results_2 = ice_results_2, clusters = clusters)
    else:
        raise Exception("Plot methods for matplotlib have not yet been implemented.")
     
def ice_clusters(ice_results, height = None, width = None, ice_results_2 = None, clusters = None):
    """Plot the cluster centers from functional clustering of ICE curves.
    
    Args:
        ice_results (ICE_Result): Result from calling the method interpretation.compute_ice with functional clustering.
        height (int, optional): Height of resulting figure. Defaults to None.
        width (int, optional): Width of resulting figures. Defaults to None.
        ice_results_2 (ICE_Result, optional): Result from an ICE computation for a different model (version). 
            Here, the result does not need to contain functional clustering.
            If specified, the ICE curves in this result are clustered according to the clustering of the other results. Defaults to None.
        clusters (iterable of int): Defines the clusters that will be plot. If None, all clusters will be plot. Default to None.

    Raises:
        Exception: If ice_results_2 was computed on a different data (version) or with different start_index, and exception is thrown.
    """
    if ice_results.cluster_centers is None:
        raise Exception('No clusters have yet been computed. Call compute_ice with clusering_param to compute clusters.')

    if has_plotly:
        _ice_clusters_plotly(ice_results, height=height, width=width, ice_results_2=ice_results_2, clusters = clusters)
    else:
        raise Exception("Plot methods for matplotlib have not yet been implemented.")
    
def ice_diff(ice_results, ice_results_2, n_curves=10, ord = 2, height = None, width = None):
    """It computes the indices of the n-th largest ICE differences between two models (w.r.t. a specified vector norm) and plots these ICE graphs.
    
    Args:
        ice_results (tools.interpretation.ICE_Results): [description]
        ice_results_2 (tools.interpretation.ICE_Results): [description]
        n_curves (int, optional): [description]. Defaults to 10.
        ord (int, optional): Defines the norm to be used to measure distance between two ICE curves, see numpy's valid strings for ord in linalg.norm [description]. Defaults to 2.
        
    """
    ice_results._validate_for_comparison(ice_results_2)
    if n_curves > ice_results.ice.shape[0]:
        raise Exception('Number of desired curves exceeds number of all curves.')
    tmp = ice_results.ice - ice_results_2.ice
    distance = np.linalg.norm(ice_results.ice - ice_results_2.ice, ord = ord, axis = 1)
    indices = [i for i in range(ice_results.ice.shape[0])]
    tmp = sorted(zip(distance, indices))
    ice_points = [tmp[i][1] for i in range(n_curves)]
    if has_plotly:
        _ice_plotly(ice_results, height=height, width=width, ice_points = ice_points, ice_results_2 = ice_results_2)
    else:
        raise Exception("Plot methods for matplotlib have not yet been implemented.")
    




import warnings
import logging
import collections
import numpy as np
from deepdiff import DeepDiff
from pailab.ml_repo.repo import MLObjectType, MLRepo, NamingConventions
from pailab.ml_repo.repo_objects import RepoInfoKey, DataSet  # pylint: disable=E0401
from pailab.ml_repo.repo_store import RepoStore  # pylint: disable=E0401
import pailab.ml_repo.repo_store as repo_store
import pailab.ml_repo.repo_objects as repo_objects
logger = logging.getLogger(__name__)

#region collections and items

   
def get_test_summary(repo:MLRepo):
    """Return test summary for all labeled models and all latest models
    
    Args:
        repo (MLRepo): repo
    """
    pass



# the followng imports are needed to hash parameters
import hashlib
import json
import functools

#region caching

class _CachedResults(repo_objects.RepoObject):
    """Object to store results of functions for caching reasons in the repo.
    
    """

    def __init__(self,repo_info, x):
        """Object to store results of functions for caching reasons in the repo.
        
        Args:
            repo_info (dict or RepoInfo object): repo info
            x (objects or tuple of objects): return values of a function to be cached
        """

        super(_CachedResults, self).__init__(repo_info)
        if isinstance(x,tuple):
            self._tuple = []
            for i in range(len(x)):
                self._add_object(x[i], i)
        else:
            if hasattr(x, 'repo_info'):
                self.repo_x = x
            else:
                self.x = x
                if isinstance(x, np.ndarray):
                    self.repo_info[RepoInfoKey.BIG_OBJECTS].append('x')

    def _add_object(self, obj, counter):
        # repo objects are added separately to the repo
        if hasattr(obj, 'repo_info'):
            name = 'repo_obj_' + str(counter)
            setattr(self, name, '')
        else:
            name = 'obj_' + str(counter)
        setattr(self, name, obj)
        self._tuple.append(name)
        if isinstance(obj, np.ndarray):
            self.repo_info[RepoInfoKey.BIG_OBJECTS].append(name)
        
    def _add_to_repo(self, ml_repo):
        if hasattr(self, '_tuple'):
            for name in self._tuple:
                if name.startswith('repo'):
                    obj = getattr(self, name)
                    ml_repo.add(obj)
                    delattr(self, name)
                    setattr(self, name +'_version', obj.repo_info.version)
                    setattr(self, name + '_name', obj.repo_info.name)
        if hasattr(self,'repo_x'):
            ml_repo.add(self.repo_x)
            delattr('repo_x')
            setattr(self,'repo_x_name', self.repo_x.repo_info.name)
            setattr(self,'repo_x_version', self.repo_x.repo_info.version)
        ml_repo.add(self)
            
    def _fill_from_repo(self, ml_repo):
        if hasattr(self,'repo_x_name'):
            obj = ml_repo.get(self.repo_x_name, self.repo_x_version)
            setattr(self, 'repo_x', obj)
            
        if hasattr(self, '_tuple'):
            for name in self._tuple:
                if name.startswith('repo'):
                    obj = ml_repo.get(getattr(self, name + '_name'), getattr(self, name + '_version'))
                    setattr(self, name, obj)
            
    def _get(self):
        if hasattr(self,'_tuple'):
            tmp = [getattr(self, name) for name in self._tuple]
            return tuple(tmp)
        if hasattr(self,'repo_x'):
            return self.repo_x
        return self.x


def cache_f(f, ml_repo, *args, **kwargs):
    """Cache results of a function in the repo.

    Caches results of a function in the repo. Here, it uses a hashkey generated as MD5 sum from the function
    arguments that are not RepoObjects (serializing them as json). For RepoObjects it uses their version number to search if the function 
    has already been executed with the given parameters. It stores the results in the repo as a CachedResult object where RepoObjects 
    contained in the CachedResults are stored separately in the repo.
    
    Args:
        f (function): function to be cached
        ml_repo (MLRepo): repository used to store the results
    
    Returns:
        the results of function f
    """

    tmp = []
    modification_info = {}
    for arg in args:
        if hasattr(arg, 'repo_info'):
            modification_info[arg.repo_info.name] = arg.repo_info.version
        else:
            tmp.append(arg)
    param = {}
    param['args'] = tmp
    param['kwargs'] ={}
    for k,v in kwargs.items():
        if hasattr(v, 'repo_info'):
            modification_info[v.repo_info.name: v.repo_info.version]
        else:
            param['kwargs'][k] = v
            
    param_hash = hashlib.md5(json.dumps(param, sort_keys=True).encode('utf-8')).hexdigest()
    modification_info['param_hash'] = param_hash
    hash_result_name = f.__name__
    result = ml_repo.get(hash_result_name, version = None, modifier_versions=modification_info, full_object=True, throw_error_not_exist = False)
    if result == []:
        x = f(*args, **kwargs)
        hash_results =  _CachedResults({RepoInfoKey.NAME: hash_result_name, RepoInfoKey.MODIFICATION_INFO: modification_info, RepoInfoKey.CATEGORY: MLObjectType.CACHED_VALUE}, x)
        hash_results._add_to_repo(ml_repo)
        return x
    else:
        result._fill_from_repo(ml_repo)
        return result._get()
    
def ml_cache(f):
    @functools.wraps(f)
    def wrapper(*argv, **kwargs):
        if 'cache' in kwargs.keys():
            ml_repo = kwargs['cache']
            if ml_repo is None:
                return f(*argv, **kwargs)    
            del kwargs['cache']
            return cache_f(f, ml_repo, *argv, **kwargs)
        else:
            return f(*argv, **kwargs)
    return wrapper

#endregion



def get_model_measure_list(ml_repo, measure_type, data, data_version = RepoStore.LAST_VERSION):
    """Return a list of models and their version together with the respective measure.
    
    Examples:
        To get a list of measures for the different models, their version and measures for measure 'mse' on the dataset 'sample1'::

            >> import pailab.tools.tools as tools
            >> tools.get_model_measure_list(ml_repo,  'mse', 'sample1')
    Args:
        ml_repo (MLRepo): MLRepo.
        measure_type (str): Name of measure type which will be returned, i.e. 'mse'. Note that the respective measure must have been added and computed before.
        data (str): Name of data on which the measure type is computed.
        data_version (str): Version of the data on which the measure has been computed.
    """
    def _get_model_info(mod_info):
        for k,v in mod_info.items():
            if k.endswith('/model'):
                return k,v

    tmp = ml_repo.get_names(MLObjectType.MEASURE)
    result = []
    missing_measure = False
    for n in tmp:
        if data in n and measure_type in n:
            candidates = ml_repo.get(n, version = None, modifier_versions = {data: data_version}, throw_error_not_exist=False)
            if not isinstance(candidates, list):
                candidates = [candidates]
            if candidates == []:
                logger.warning('Missing measure ' + n + ' for data version: ' + data_version)
                missing_measure = True
            for c in candidates:
                model_name, model_version = _get_model_info(c.repo_info.modification_info)
                result.append({'model': model_name, 'version': model_version, measure_type + ', ' + data: c.value})
    if missing_measure:
        warnings.warn('There were measures missing for the given data version, see log for details.')
    return result
       
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeRegressor
from sklearn.metrics import mean_squared_error
from sklearn.cluster import KMeans

class ModelAnalyzer:
    def __init__(self, ml_repo):
        self._ml_repo = ml_repo
        self._decision_tree = None
        self.result = {}

    @staticmethod        
    def _compute_local_model_coeffs(x_data, model_eval_function, model, y_coordinate,  n_samples, factor):
        rnd_shape = (n_samples,) + x_data.shape[1:]
        np.random.seed(42)
        rand_values = (2.0-2.0*factor)*np.random.random_sample(rnd_shape)+factor
        local_model_coeff = np.empty(x_data.shape)
        mse = np.empty((x_data.shape[0],))
        training_data_x = np.empty(rnd_shape)
        for i in range(x_data.shape[0]):
            for j in range(rand_values.shape[0]):
                training_data_x[j,:] = np.multiply(rand_values[j,:], x_data[i,:])
            training_data_y = model_eval_function.create()(model, training_data_x)[:,y_coordinate]
            reg = LinearRegression().fit(rand_values, training_data_y)
            prediction = reg.predict(rand_values)
            mse[i] = mean_squared_error(training_data_y, prediction)
            local_model_coeff[i,:] = reg.coef_

        return local_model_coeff, mse

    @staticmethod
    def _get_tree_figures(decision_tree, data, mse):
        #first extract tree leafs for statistics (statistics only on leafs)
        leaf_nodes = {}
        num_elements_per_node = [0]*decision_tree.tree_.node_count
        for i in range(len(decision_tree.tree_.children_left)):
            if decision_tree.tree_.children_left[i] == decision_tree.tree_.children_right[i]:
                leaf_nodes[i] = {'mse_max':-1.0, 'mse_min': 20000000000.0, 'mse_mean': 0.0}
        #depp = np.unique(decision_tree.tree_.apply(x.astype(np.float32)))
        tmp = decision_tree.apply(data)

        for i in range(len(tmp)):
            num_elements_per_node[tmp[i]] += 1
        for i in range(mse.shape[0]):
            leaf_nodes[tmp[i]]['mse_max'] = max(leaf_nodes[tmp[i]]['mse_max'], mse[i])
            leaf_nodes[tmp[i]]['mse_min'] = min(leaf_nodes[tmp[i]]['mse_min'], mse[i])
            leaf_nodes[tmp[i]]['mse_mean'] +=  mse[i]
        for k,v in leaf_nodes.items():
            v['num_data_points'] = num_elements_per_node[k]
            v['mse_mean'] /= float(num_elements_per_node[k])
            v['model_coefficients'] = decision_tree.tree_.value[k][:,0].tolist()
        return leaf_nodes, tmp

    
    @staticmethod
    def _create_result(name, model, data, param, result_data, big_data = None):
        result = repo_objects.Result(result_data, big_data)
        result.repo_info.name = name
        logger.debug('Create md5 hash for parameters.')
        param_hash = hashlib.md5(json.dumps(param, sort_keys=True).encode('utf-8')).hexdigest()
        logger.debug('Add md5 hash for parameters to modification_info of results.')
        result.repo_info.modification_info = {model.repo_info.name: model.repo_info.version, 
                    data.repo_info.name: data.repo_info.version, 'param_hash': param_hash}
        return result


    def analyze_local_model(self, model, data_str, n_samples, version = RepoStore.LAST_VERSION, data_version = RepoStore.LAST_VERSION, 
                    y_coordinate=None, start_index = 0, end_index= 100, full_object = True, factor=0.1, 
                    max_depth = 4):
        logger.info('Start analyzing local model.')
        if y_coordinate is None:
            y_coordinate = 0
        if isinstance(y_coordinate, str):
            raise NotImplementedError()
        param = {'max_depth': max_depth, 'factor': factor, 'n_samples': n_samples, 
            'y_coordinate': y_coordinate, 'start_index': start_index, 'end_index': end_index}
        param_hash = hashlib.md5(json.dumps(param, sort_keys=True).encode('utf-8')).hexdigest()

        # check if results of analysis are already stored in the repo
        model_ = self._ml_repo.get(model, version, full_object = False)
        data_ = self._ml_repo.get(data_str, data_version, full_object=False)
        result_name = 'analyzer_local_model_' + model_.repo_info.name + '_' + data_.repo_info.name
        result = self._ml_repo.get(result_name, None, 
                modifier_versions={model_.repo_info.name: model_.repo_info.version, data_.repo_info.name: data_.repo_info.version, 'param_hash': param_hash},
                throw_error_not_exist=False, full_object= full_object)
        if result != []:
            if isinstance(result,list):
                return result[0]
            else:
                return result

        model_definition_name = model.split('/')[0]
        model_name = model
        model = self._ml_repo.get(model, version, full_object = True)
        model_def_version = model.repo_info[RepoInfoKey.MODIFICATION_INFO][model_definition_name]
        model_definition = self._ml_repo.get(model_definition_name, model_def_version)
        data = self._ml_repo.get(data_str, data_version, full_object=True)
        eval_func = self._ml_repo.get(model_definition.eval_function, RepoStore.LAST_VERSION)
        
            
        data_ = data.x_data[start_index:end_index, :]
        local_model_coeff, mse = ModelAnalyzer._compute_local_model_coeffs(data_, eval_func, model, y_coordinate, n_samples, factor)
        self._decision_tree = DecisionTreeRegressor(max_depth=max_depth)
        self._decision_tree.fit(data_, local_model_coeff)
        
        self.result['node_statistics'], datapoint_to_leaf_node_idx = ModelAnalyzer._get_tree_figures(self._decision_tree, data_, mse)
        # store result in repo
        self.result['parameter'] = param
        self.result['data'] = data_str
        self.result['model'] = model_name
        self.result['x_coord_names'] = data.x_coord_names
        result = ModelAnalyzer._create_result(result_name, 
            model, data, param, self.result, {'local_model_coeff':local_model_coeff, 'data_to_leaf_index':datapoint_to_leaf_node_idx, 'mse': mse })
        self._ml_repo.add(result)
        return result

   
    





class RepoStoreFactory:
    """ the repo stores factory

    The factory is used to create Repo handlers

    """

    @staticmethod
    def get_repo_stores():
        """ return a list of supported repo handlers

        Returns:
            list of str -- the list of supported handlers
        """

        return ['disk_handler', 'git_handler', 'memory_handler']

    @staticmethod
    def get(repo_store_type, **kwargs):
        """ this method returns a handler of the specified type

        This method is used to construct handler of the specified type using the provided arguments.
        Currently these types are supported:
        * disk_handler
        * git_handler
        * memory_handler

        Args:
            repo_store_type (str): the name of the repo store type

        Raises:
            Exception: raises an exception if the type is not supported

        Returns:
            RepoStore -- the constructed handler
        """

        if repo_store_type == 'disk_handler':
            from pailab.ml_repo.disk_handler import RepoObjectDiskStorage
            return RepoObjectDiskStorage(**kwargs)
        elif repo_store_type == 'git_handler':
            from pailab.ml_repo.git_handler import RepoObjectGitStorage
            return RepoObjectGitStorage(**kwargs)
        elif repo_store_type == 'memory_handler':
            from pailab.ml_repo.memory_handler import RepoObjectMemoryStorage
            return RepoObjectMemoryStorage()
        raise Exception('Cannot create RepoStore: Unknown repo type ' + repo_store_type +
                        '. Use only types from the list returned by RepoStoreFactory.get_repo_stores().')


class NumpyStoreFactory:
    """ class to construct a numpy store for big data
    """

    @staticmethod
    def get_numpy_stores():
        """ this method returns the supported types

        Returns:
            list of str -- list of supported big data handler types
        """

        return ['memory_handler', 'hdf_handler', 'hdf_remote_handler']

    @staticmethod
    def get(numpy_store_type, **kwargs):
        """ this method returns a handler of the specified type for handling big data

        This method is used to construct a big data handler of the specified type using the provided arguments.
        Currently these types are supported:
        * memory_handler
        * hdf_handler

        Args:
            numpy_store_type (str): the name of the big data store type

        Raises:
            Exception: raises an exception if the type is not supported

        Returns:
            NumpyStore -- the constructed handler
        """
        if numpy_store_type == 'memory_handler':
            from pailab.ml_repo.memory_handler import NumpyMemoryStorage
            return NumpyMemoryStorage(**kwargs)
        elif numpy_store_type == 'hdf_handler':
            from pailab.ml_repo.numpy_handler_hdf import NumpyHDFStorage
            return NumpyHDFStorage(**kwargs)
        elif numpy_store_type == 'hdf_remote_handler':
            from pailab.ml_repo.numpy_handler_hdf import NumpyHDFRemoteStorage
            return NumpyHDFRemoteStorage(**kwargs)
        raise Exception('Cannot create NumpyStore: Unknown  type ' + numpy_store_type +
                        '. Use only types from the list returned by NumpyStoreFactory.get_numpy_stores().')



import uuid
import datetime

import abc
from pailab.ml_repo.repo_objects import RepoInfoKey  # pylint: disable=E0401


def _version_str():
    """ returns the version as string

    Returns:
        str -- a new version
    """

    return str(uuid.uuid1())


def _time_from_version(v):
    """ Return the time included in uuid

    Args:
        v (str): string representin the uuid

    Returns:
        datetime -- the datetime
    """

    if isinstance(v, str):
        tmp = uuid.UUID(v)
        return datetime.datetime(1582, 10, 15) + datetime.timedelta(microseconds=tmp.time / 10)
    return datetime.datetime(1582, 10, 15) + datetime.timedelta(microseconds=v.time / 10)


FIRST_VERSION = 'first'
LAST_VERSION = 'last'


class RepoScriptStore(abc.ABC):
    @abc.abstractmethod
    def add(self, script_file):
        """ Add a script to the storage.


        Args:
            script_file (string): file (incluing path) of script

        """

        pass

    @abc.abstractmethod
    def get(self, name, versions=None):
        """

        """
        pass


class RepoStore(abc.ABC):

    FIRST_VERSION = FIRST_VERSION
    LAST_VERSION = LAST_VERSION

    """ Abstract base class for all storages which can be used in the ML repository

    """

    def _replace_version_placeholder(self, name, versions, throw_error_not_exist=True):
        """ replaces the version placeholder

        Args:
            name (str): identifier of the object
            versions (str): the version identifier
            throw_error_not_exist (bool): throw an error if not exists. Defaults to True.

        Returns:
            [type] -- list of versions
        """

        def replace_version(name, version, throw_error_not_exist):
            if version == FIRST_VERSION:
                return self.get_first_version(name, throw_error_not_exist)
            if version == LAST_VERSION:
                return self.get_latest_version(name, throw_error_not_exist)
            if isinstance(version, int):
                return self.get_version(name, version, throw_error_not_exist)
            return version
        if isinstance(versions, str):
            versions = replace_version(name, versions, throw_error_not_exist)
        if isinstance(versions, tuple):
            versions = (replace_version(name, versions[0], throw_error_not_exist),
                        replace_version(name, versions[1], throw_error_not_exist))
        if isinstance(versions, list):
            versions = [replace_version(
                name, v, throw_error_not_exist) for v in versions]
        return versions

    @abc.abstractmethod
    def _add(self, obj):
        """ Add an object to the storage.

        This method is called internally by the method add that is setting a version if there has not version already been set.

        Args:
            obj (RepoObject): repository object

        Raises:
            Exception if an object with same name already exists.
        """
        pass

    def add(self, obj):
        """ Add an object to the storage.


        Args:
            obj (RepoObject|list(RepoObject)): repository object or list o repository objects

        Raises:
            Exception if an object with same name already exists.
        """
        if obj['repo_info'][RepoInfoKey.VERSION.value] is None:
            obj['repo_info'][RepoInfoKey.VERSION.value] = _version_str()
        self._add(obj)
        return obj['repo_info'][RepoInfoKey.VERSION.value]

    @abc.abstractmethod
    def replace(self, obj):
        """ Overwrite existing object without incrementing version

        Args:
            obj (RepoObject): repo object to be overwritten
        """

        pass

    def _get_by_modification_info(self, modifier_name, modifier_version, object_types=[]):
        """ Return list of all objects which were modified by a given object.

        This method may be overwritten by subclasses to enhance performance.

        Args:
            modifier_name (str): name of object which modified th searched objects
            modifier_version (str): version of object which modified th searched objects
            object_types (list of str): list of strings defining the object types. Defaults to [].

        Returns:
            list -- list of objects, empty if no such objects exist
        """

        result = []
        modifier = {modifier_name: modifier_version}
        for category in object_types:
            names = self.get_names(category)
            for n in names:
                objs = self.get(n, modifier_versions=modifier,
                                throw_error_not_exist=False, throw_error_not_unique=False)
                if isinstance(objs, list):
                    result.extend(objs)
                else:
                    result.append(objs)
        return result

    def get(self, name, versions=None, modifier_versions=None, obj_fields=None,  repo_info_fields=None,
            throw_error_not_exist=True, throw_error_not_unique=True):
        """ Get a dictionary/list of dictionaries fulffilling the conditions.

            Returns a list of objects matching the name and whose
                -version is in the given list of versions
                -modifiers match the version number/are in the list of version numbers of the given modifiers
        Args:
            name (str): object id
            versions (list, version_number, tuple): either a list of versions or a single version of the objects to be returned,. Defaults to None.
                    if None, the condition on version is ignored. If a tuple is given, this tuple defines a version intervall, i.e.
                    all versions between the first and last entry (both including) are returned. In addition FIRST_VERSION and LAST_VERSION can be used for versions to access
                    the last/first version.
            modifier_versions (dictionary): modifier ids together with version specs which are matched by the returned object.. Defaults to None.
            obj_fields (list of str or str): list of strings identifying the fields which will be returned in the dictionary,
                                                        if None, no fields are returned, if set to 'all', all fields will be returned . Defaults to None.
            repo_info_fields (list of str or str): list of strings identifying the fields of the repo_info dict which will be returned in the dictionary,
                                                        if None, no fields are returned, if set to 'all', all fields will be returned. Defaults to None.
            throw_error_not_exist (bool): true - throw error if not exists, else return []. Defaults to True.
            throw_error_not_unique (bool): true - throw error if item is not unique, else return []. Defaults to True.

        Returns:
            RepoObject or list thereof -- The repo object
        """

        versions = self._replace_version_placeholder(
            name, versions, throw_error_not_exist)
        if modifier_versions is not None:
            for k, v in modifier_versions.items():
                modifier_versions[k] = self._replace_version_placeholder(
                    k, v, throw_error_not_exist)
        return self._get(name, versions, modifier_versions,
                         obj_fields, repo_info_fields,
                         throw_error_not_exist, throw_error_not_unique)

    def push(self):
        """ Push changes to an external repo.
        """

        pass

    def pull(self):
        """ Pull changes from an external repo
        """

        pass

    @abc.abstractmethod
    def _delete(self, name, version):
        """ Delete an object with a predefined version

        Args:
            name (str): identifier of the object
            version (str): object version
        """

        pass

    @abc.abstractmethod
    def _get(self, name, versions=None, modifier_versions=None, obj_fields=None,  repo_info_fields=None,
             throw_error_not_exist=True, throw_error_not_unique=True):
        """ Get a dictionary/list of dictionaries fulffilling the conditions.

            Returns a list of objects matching the name and whose
                -version is in the given list of versions
                -modifiers match the version number/are in the list of version numbers of the given modifiers
        Args:
            name (str): object id
            versions (list, version_number, tuple): either a list of versions or a single version of the objects to be returned,. Defaults to None.
                    if None, the condition on version is ignored. If a tuple is given, this tuple defines a version intervall, i.e.
                    all versions between the first and last entry (both including) are returned. In addition FIRST_VERSION and LAST_VERSION can be used for versions to access
                    the last/first version.
            modifier_versions (dictionary): modifier ids together with version specs which are matched by the returned object.. Defaults to None.
            obj_fields (list of str or str): list of strings identifying the fields which will be returned in the dictionary,
                                                        if None, no fields are returned, if set to 'all', all fields will be returned . Defaults to None.
            repo_info_fields (list of str or str): list of strings identifying the fields of the repo_info dict which will be returned in the dictionary,
                                                        if None, no fields are returned, if set to 'all', all fields will be returned. Defaults to None.
            throw_error_not_exist (bool): true - throw error if not exists, else return []. Defaults to True.
            throw_error_not_unique (bool): true - throw error if item is not unique, else return []. Defaults to True.

        Returns:
            RepoObject or list thereof -- The repo object
        """
        pass

    @abc.abstractmethod
    def get_names(self, ml_obj_type):
        """ Return the names of all objects belonging to the given category.

        Args:
            ml_obj_type (str): Value of MLObjectType-Enum specifying the category for which all names will be returned
        """
        pass

    @abc.abstractmethod
    def get_version(self, name, offset, throw_error_not_exist=True):
        """ Return versionnumber for the given offset

        If offset >= 0 it returns the version number of the offset version, if <0 it returns according to the
        python list logic the version number of the (offset-1)-last version

        Args:
            name (str): name of object
            offset (int): offset
            throw_error_not_exist (bool): true - throw error if not exists, else return []. Defaults to True.

        """
        pass

    @abc.abstractmethod
    def get_latest_version(self, name, throw_error_not_exist=True):
        """ Return latest version number of object in the storage

        Args:
            name (str): object name
            throw_error_not_exist (bool): true - throw error if not exists, else return []. Defaults to True.

        Returns:
            str -- latest version number
        """

        return self.get(name, versions=RepoStore.LAST_VERSION, repo_info_fields=[RepoInfoKey.VERSION])[0]['repo_info'][RepoInfoKey.VERSION.value]

    @abc.abstractmethod
    def get_first_version(self, name, throw_error_not_exist=True):
        """ Return version number of first (in a temporal sense) object in storage

        Args:
            name (str): object name for which the version is returned
            throw_error_not_exist (bool): true - throw error if not exists, else return []. Defaults to True.

        Raises:
            NotImplementedError -- [description]
        """

        raise NotImplementedError()

    def object_exists(self, name, version=LAST_VERSION):
        """ Returns True if an object with the given name and version exists.

        Args:
            name (string): object name
            version (version number): version number. Defaults to LAST_VERSION.
        """
        obj = []
        try:
            obj = self.get(name, versions=[version],
                           throw_error_not_exist=False, throw_error_not_unique=False)
        except:
            pass
        return len(obj) != 0


class NumpyStore(abc.ABC):
    """ class to handle big objects
    """

    @abc.abstractmethod
    def _delete(self, name, version):
        """ Delete an object with a predefined version.

        This method is only internally used by the MLRepo.

        Args:
            name (str): name of object
            version (str): object version
        """

        pass

    @abc.abstractmethod
    def add(self, name, version, numpy_dict):
        """ Add numpy data from an object to the storage.

        Args:
            name (str): Name (as string) of object
            version (str): object version
            numpy_dict (numpy dict): numpy dictionary
        """

        pass

    @abc.abstractmethod
    def append(self, name, version_old, version_new, numpy_dict):
        """ Append data to an existing object

        Args:
            name (str): name of data object to be returned
            version_old (str): version of the object where the data will be appended
            version_new (str): version of the new objct after appending the data
            numpy_dict (dict): dictionary containing the values
        """

        pass

    @abc.abstractmethod
    def get(self, name, version, from_index=0, to_index=None):
        """ get the numpy object for a name and a version, rows can be used

        Args:
            name (str): identifier of the object
            version (str): version of the object
            from_index (int): the index from which the data should be taken. Defaults to 0.
            to_index (int or None): the index to which the data is returned (None means till the end). Defaults to None.

        Returns:
            numpy array -- the numpy object to return
        """

    def push(self):
        """ Push changes to an external repo.
        """

        pass

    def pull(self):
        """ Pull changes from an external repo
        """

        pass



"""Main entry point for pailab command line """
import argparse
import sys
import getpass
from pailab import MLRepo
#from pailab import main


def _get_default_config():
    return {'user': 'test_user',
            'workspace': 'c:/dummy',
            'repo_store':
            {
                'type': 'git_handler',
                'config': {
                    'folder': 'c:/dummy',
                    'file_format': 'json',
                    'remote': 'https://github.com/pailabteam/scott_chesney.git'
                }
            },
            'numpy_store':
            {
                'type': 'hdf_remote_handler',
                'config': {
                    'folder': 'c:/dummy',
                    'remote_store': {
                        'type': 'gcs',
                        'config': {
                            'bucket': 'dummy_bucket',
                            'project': 'dummy_project'
                        }
                    },
                    'sync_get': False,
                    'sync_add': True
                }
            },
            'job_runner':
            {
                'type': 'simple',
                'config': {
                    'throw_job_error': True
                }
            }
            }


def _set_numpy_store_config(argument, config):
    """[summary]

    Examples:

        hdf_remote_handler@gcs:project:bucket
    Args:
        args ([type]): [description]
    """
    result = {}
    tmp = argument.split('@')
    store_type = tmp[0]
    config['numpy_store']['type'] = store_type
    if store_type == 'hdf_remote_handler':
        if len(tmp) < 2:
            raise Excetion(
                'Please specify a remote, e.g. hdf_remote_handler@remote_settings.')
        tmp = tmp[1].split(':')
        if tmp[0] == 'gcs':
            if len(tmp) < 3:
                raise Exception(
                    'Please specify project and bucket for gcs, e.g. hdf_remote_handler@gcs:project:bucket.')
            config['numpy_store']['config']['remote_store']['type'] = 'gcs'
            config['numpy_store']['config']['remote_store']['config']['project'] = tmp[1]
            config['numpy_store']['config']['remote_store']['config']['bucket'] = tmp[2]
            return
        raise Exception('Unknown remote type ' + tmp[0])
    raise Exception('Unknown store type ' + store_type)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='')
    parser.add_argument('command', help='one of the command(s) create|init')
    parser.add_argument('workspace', help='directory for the workspace')
    parser.add_argument(
        'obj_handler', type=str, help='definition of store for the repo objects')
    parser.add_argument(
        'bigobj_handler', type=str, help='definition of store for the big object parts of the repo objects')
    parser.add_argument('-u', '--user', type=str,
                        help='username used internally, if not specified, the operating system user is used')
    args = parser.parse_args()
    default_config = _get_default_config()
    default_config['workspace'] = args.workspace
    if args.user:
        default_config['user'] = args.user
    else:
        default_config['user'] = getpass.getuser()
    default_config['numpy_store']['config']['folder'] = args.workspace + '/hdf'
    default_config['repo_store']['config']['folder'] = args.workspace + '/objects'
    _set_numpy_store_config(args.bigobj_handler, default_config)
    tmp = args.obj_handler.split('@')
    default_config['repo_store']['type'] = tmp[0]
    if tmp[0] == 'git_handler':
        if len(tmp) > 1:
            default_config['repo_store']['config']['remote'] = tmp[1]

    #print('Creating repo with config: ' + str(default_config))
    ml_repo = MLRepo(config=default_config, save_config=True)



import os
import logging
from google.cloud import storage
logger = logging.getLogger(__name__)


class RemoteGCS:
    def __init__(self, bucket='', project=None, credentials=None):
        self._storage_client = storage.Client(
            project=project, credentials=credentials)
        self._bucket = self._storage_client.get_bucket(bucket)
        self._bucket_name = bucket
        self._project = project

    def _get_bucket(self):
        return self._bucket

    def _remote_file_list(self):
        return [x.name for x in self._get_bucket().list_blobs()]

    def _download_file(self, local_filename, remote_filename):
        bucket = self._get_bucket()
        remote_filename = remote_filename.replace('\\', '/')
        blob = bucket.blob(remote_filename)
        # if not os.path.exists(local_filename):
        #    os.path.ma
        path_to_local_file = os.path.dirname(local_filename)
        if not os.path.exists(path_to_local_file):
            os.makedirs(path_to_local_file)
        with open(local_filename, 'wb') as file_obj:
            logger.debug('Start downloading ' +
                         remote_filename + ' to ' + local_filename)
            blob.download_to_file(file_obj)
            logger.debug('Finished downloading ' + remote_filename)

    def _upload_file(self,  local_filename, remote_filename):
        bucket = self._get_bucket()
        remote_filename = remote_filename.replace('\\', '/')
        blob = bucket.blob(remote_filename)
        with open(local_filename, 'rb') as file_obj:
            logger.debug('Start uploading file ' +
                         local_filename + ' to ' + remote_filename)
            blob.upload_from_file(file_obj)
            logger.debug('Start uploading file ' + local_filename)

    def file_exists(self, filename):
        filename = filename.replace('\\', '/')
        return storage.Blob(bucket=self._bucket, name=filename).exists(self._storage_client)





import os
from contextlib import closing
import sqlite3
from datetime import datetime, timedelta

import pathlib
import pailab.ml_repo.repo_objects as repo_objects
from pailab.ml_repo.repo_store import RepoInfoKey, _time_from_version
import pailab.ml_repo.repo as repo
from pailab.ml_repo.repo_store import RepoStore
from shutil import copy
import logging
logger = logging.getLogger(__name__)
logger_sql = logging.getLogger(__name__ + '_SQLITE')



class RepoObjectDiskStorage(RepoStore):
    """ The RepoObjectDiskStorage class
    """

    # region private

    # region json encoding for enums
    PUBLIC_ENUMS = {
        'MLObjectType': repo.MLObjectType,
        # ...
    }

    # endregion

    def _sqlite_db_name(self):
        """ return the sqlite db path and file name
        
        Returns:
            str -- return the path and file of the sqlite db
        """

        return self._main_dir + '/.version.sqlite'

    def _create_new_db(self):
        """ Creates a new sqlite db
        """

        self._conn = sqlite3.connect(self._sqlite_db_name())
        # three tables: one with category->name mapping, one with category, version, and one with modification info
        # mapping
        with closing(self._conn.cursor()) as cursor:
            try:
                logger.info('Executing')
                cursor.execute(
                        '''CREATE TABLE mapping (name text PRIMARY KEY, category text)''')
                # versions
                cursor.execute(
                        '''CREATE TABLE versions (name TEXT NOT NULL, version TEXT NOT NULL, path TEXT NOT NULL, file TEXT NOT NULL, uuid_time TIMESTAMP,
                                            insert_time TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL, PRIMARY KEY(name, version) ) ''')

                # modification_info
                cursor.execute(
                        '''CREATE TABLE modification_info (name TEXT NOT NULL, version TEXT NOT NULL, modifier TEXT NOT NULL, modifier_version TEXT NOT NULL, modifier_uuid_time TIMESTAMP, PRIMARY KEY(name, version, modifier) ) ''')
                self._conn.commit()
            except:
                logger.error(
                    'An error occured during creation of new db, rolling back.')
                self._conn.rollback()
        
    def _setup_new(self):
        """ Setup of the handler
        """

        if not os.path.exists(self._main_dir):
            os.makedirs(self._main_dir)
        if not os.path.exists(self._sqlite_db_name()):
            self._create_new_db()
        else:
            self._conn = sqlite3.connect(self._sqlite_db_name())
        self._conn.set_trace_callback(logger_sql.info)
    # endregion

    def __init__(self, folder, file_format='pickle'):
        """ Constructor
        
        Args:
            folder (str):directory used to store the objects in files as well as the sqlite database
            file_format (str):the fileformat to save. Defaults to 'pickle'.
        
        Raises:
            Exception: raise an exception if the file format is unknown (currently only pickle and json is supported)
        
        Returns:
            [type] -- [description]
        """

        self._main_dir = folder
        self._setup_new()
        self._file_format = file_format
        self._extension = '.pck'
        if self._file_format == 'pickle':
            import pickle

            def __pickle_save(file_prefix, obj):
                with open(file_prefix + '.pck', 'wb') as f:
                    pickle.dump(obj, f)

            def __pickle_load(file_prefix):
                with open(file_prefix+'.pck', 'rb') as f:
                    return pickle.load(f)

            self._save_function = __pickle_save
            self._load_function = __pickle_load
        elif self._file_format == 'json':
            import json

            class CustomEncoder(json.JSONEncoder):
                def default(self, obj):
                    if type(obj) in RepoObjectDiskStorage.PUBLIC_ENUMS.values():
                        return {"__enum__": str(obj)}
                    else:
                        if isinstance(obj, datetime):
                            return {"__datetime__": str(obj)}
                    return json.JSONEncoder.default(self, obj)

            class CustomDecoder(json.JSONDecoder):
                def __init__(self, *args, **kwargs):
                    json.JSONDecoder.__init__(
                        self, object_hook=self.object_hook, *args, **kwargs)

                def object_hook(self, obj):
                    if "__enum__" in obj:
                        name, member = obj["__enum__"].split(".")
                        return getattr(RepoObjectDiskStorage.PUBLIC_ENUMS[name], member)
                    if "__datetime__" in obj:
                        return datetime.strptime(obj["__datetime__"], 'YYYY-MM-DD HH:MM:SS.mmmmmm ')
                    return obj

            def __json_load(file_prefix):
                with open(file_prefix+'.json', 'r') as f:
                    return json.load(f, cls=CustomDecoder)

            def __json_save(file_prefix, obj):
                with open(file_prefix + '.json', 'w') as f:
                    json.dump(obj, f, cls=CustomEncoder,
                              indent=4, separators=(',', ': '))

            self._save_function = __json_save
            self._load_function = __json_load
            self._extension = '.json'
        else:
            raise Exception("Unknown file format " + file_format)

    def _delete(self, name, version):
        """ Deletes a object 
        
        Args:
            name (str):the identifier of the object
            version (str):the version of the object
        
        Raises:
            Exception: an exception is raised if the object does not exists
        """

        condition = " where name='"+  name + "' and version='" + version + "'"
        delete_statement = "delete from modification_info " + condition
        with closing(self._conn.cursor()) as cursor:
            cursor.execute(delete_statement)
            select_statement = "select path, file from versions " + condition
            files = [row[0] + '/' + row[1]
                    for row in cursor.execute(select_statement)]
            if len(files) == 0:
                raise Exception('Deletion failed: Object ' + name + " with version " + version +' does not exist.')
            for filename in files:
                os.remove(self._main_dir + '/' + filename + self._extension)
            delete_statement = "delete from versions " + condition
            cursor.execute(delete_statement)
            #if there is no object with this name anymore, we have to remove it from mapping
            select_statement = "select path, file from versions where name='"+  name + "'"
            depp = [0 for r in cursor.execute(select_statement)]
            if len(depp) == 0:
                delete_statement = "delete from mapping where name='" + name + "'"
                cursor.execute(delete_statement)        
            self._conn.commit()
        
    def get_config(self):
        """ return the configuration
        
        Returns:
            dict -- a dictionary of the configuration
        """

        return {'folder': self._main_dir, 'file_format': self._file_format}

    def get_names(self, ml_obj_type):
        """ Return the names of all objects belonging to the given category.
        
        Args:
            ml_obj_type (str):Value of MLObjectType-Enum specifying the category for which all names will be returned
        
        Returns:
            list of str -- a list of all objects in the category
        """
        with closing(self._conn.cursor()) as cursor: 
            result = []
            for row in cursor.execute("select name, category from  mapping where category = '" + ml_obj_type + "'"):
                result.append(row[0])
            return result

    def _add(self, obj):
        """Add an object to the storage.

        Args:
            obj (RepoObject):repository object

        Raises:
            Exception if an object with same name already exists.
        """
        with self._conn:
            with closing(self._conn.cursor()) as cursor:    
                try:
                    uid_time = _time_from_version(
                        obj['repo_info'][repo_objects.RepoInfoKey.VERSION.value])
                    name = obj['repo_info'][repo_objects.RepoInfoKey.NAME.value]
                    if isinstance(obj['repo_info'][repo_objects.RepoInfoKey.CATEGORY.value], repo.MLObjectType):
                        category = obj['repo_info'][repo_objects.RepoInfoKey.CATEGORY.value].name
                    else:
                        category = obj['repo_info'][repo_objects.RepoInfoKey.CATEGORY.value]
                    exists = False
                    # region write mapping
                    for row in cursor.execute('select * from mapping where name = ' + "'" + name + "'"):
                        exists = True
                        break
                    if not exists:
                        cursor.execute(
                                "insert into mapping (name, category) VALUES ('" + name + "', '" + category + "')")
                    # endregion
                    # region write file info
                    version = obj['repo_info'][repo_objects.RepoInfoKey.VERSION.value]
                    file_sub_dir = category + '/' + name + '/'
                    os.makedirs(self._main_dir + '/' + file_sub_dir, exist_ok=True)
                    filename = version
                    cursor.execute("insert into versions (name, version, path, file, uuid_time) VALUES('" +
                            name + "', '" + version + "','" + file_sub_dir + "','" + filename + "','" + str(uid_time) + "')")
                    # endregion
                    # region write modification info
                    if repo_objects.RepoInfoKey.MODIFICATION_INFO.value in obj['repo_info']:
                        for k, v in obj['repo_info'][repo_objects.RepoInfoKey.MODIFICATION_INFO.value].items():
                            tmp = _time_from_version(v)
                            cursor.execute("insert into modification_info (name, version, modifier, modifier_version, modifier_uuid_time) VALUES ('"
                                    + name + "','" + version + "','" + k + "','" + str(v) + "','" + str(tmp) + "')")
                    # endregion
                    self._conn.commit()
                    # region write file
                    logger.debug(
                        'Write object as json file with filename ' + filename)

                    self._save_function(self._main_dir + '/' +
                                        file_sub_dir + '/' + filename, obj)
                    # endregion
                except Exception as e:
                    logger.error('Error: ' + str(e) + ', rolling back changes.')
                    self._conn.rollback()

    def get_version_condition(self, name, versions, version_column, time_column):
        """ returns the condition part of the versions for the sql statement
        
        Args:
            name (str):not used
            versions (str or list of str):a or the versions to condition on
            version_column (str):version column name
            time_column (str):time column name
        
        Returns:
            str -- the condition for the versions
        """

        version_condition = ''
        if versions is not None:
            version_condition = ' and '
        if isinstance(versions, str):
            version_condition += version_column + " = '" + versions + "'"
        else:
            if isinstance(versions, tuple):
                start_time = _time_from_version(versions[0])
                end_time = _time_from_version(versions[1])
                version_condition += "'" + str(
                    start_time) + "' <= " + time_column + " and " + time_column + "<= '" + str(end_time) + "'"
            else:
                if isinstance(versions, list):
                    version_condition += version_column + ' in ('
                    tmp = "','"
                    tmp = tmp.join([v for v in versions])
                    version_condition += "'" + tmp + "')"
        return version_condition

    def _get(self, name, versions=None, modifier_versions=None, obj_fields=None,  repo_info_fields=None,
             throw_error_not_exist=True, throw_error_not_unique=True):
        """ Get a dictionary/list of dictionaries fulffilling the conditions.

            Returns a list of objects matching the name and whose
                -version is in the given list of versions
                -modifiers match the version number/are in the list of version numbers of the given modifiers
        Args:
            name (string):object id
            versions (list, version_number, tuple):either a list of versions or a single version of the objects to be returned,. Defaults to None.
                    if None, the condition on version is ignored. If a tuple is given, this tuple defines a version intervall, i.e.
                    all versions between the first and last entry (both including) are returned. In addition FIRST_VERSION and LAST_VERSION can be used for versions to access
                    the last/first version.
            modifier_versions (dictionary):modifier ids together with version specs which are matched by the returned object.. Defaults to None.
            obj_fields (list of strings or string):list of strings identifying the fields which will be returned in the dictionary,
                                                        if None, no fields are returned, if set to 'all', all fields will be returned . Defaults to None.
            repo_info_fields (list of strings or string):list of strings identifying the fields of the repo_info dict which will be returned in the dictionary,
                                                        if None, no fields are returned, if set to 'all', all fields will be returned. Defaults to None.
        """

        category = None
        with closing(self._conn.cursor()) as cursor:
            for row in cursor.execute('select category from mapping where name = ' + "'" + name + "'"):
                category = repo.MLObjectType(row[0])
            if category is None:
                if throw_error_not_exist:
                    logger.error('no object ' + name + ' in storage.')
                    raise Exception('no object ' + name + ' in storage.')
                else:
                    return []

            version_condition = self.get_version_condition(
                name, versions, 'version', 'uuid_time')

            select_statement = "select path, file from versions where name = '" + \
                name + "'" + version_condition
            if modifier_versions is not None:
                for k, v in modifier_versions.items():
                    tmp = self.get_version_condition(
                        k, v, 'modifier_version', 'modifier_uuid_time')
                    if tmp != '':
                        select_statement += " and version in ( select version from modification_info where name ='" + \
                            name + "' and modifier = '" + k + "'" + tmp + ")"
            files = [row[0] + '/' + row[1]
                    for row in cursor.execute(select_statement)]
            objects = []
            for filename in files:
                objects.append(self._load_function(
                    self._main_dir + '/' + filename))
            self._conn.commit()
        return objects

    def get_latest_version(self, name, throw_error_not_exist=True):
        """ Determine the latest version of the object
        
        Args:
            name (str):identifier of the object
            throw_error_not_exist (bool):true - throw error if not exists, else return []. Defaults to True.
        
        Raises:
            Exception: Raises an exception if the object does not exists
        
        Returns:
            str -- the latest version string of the object
        """

        with closing(self._conn.cursor()) as cursor:
            for row in cursor.execute("select version from versions where name = '" + name + "' order by uuid_time DESC LIMIT 1"):
                result = row[0]
                return result
        if throw_error_not_exist:
            logger.error('No object with name ' + name + ' exists.')
            raise Exception('No object with name ' + name + ' exists.')
        else:
            return []

    def get_first_version(self, name, throw_error_not_exist=True):
        """ Determine the first version of the object
        
        Args:
            name (str):identifier of the object
            throw_error_not_exist (bool):true - throw error if not exists, else return []. Defaults to True.
        
        Raises:
            Exception: Raises an exception if the object does not exists
        
        Returns:
            str -- the first version string of the object
        """
        with closing(self._conn.cursor()) as cursor:
            for row in cursor.execute("select version from versions where name = '" + name + "' order by uuid_time ASC LIMIT 1"):
                result = row[0]
                return result
        if throw_error_not_exist:
            logger.error('No object with name ' + name + ' exists.')
            raise Exception('No object with name ' + name + ' exists.')
        else:
            return []

    def get_version(self, name, offset, throw_error_not_exist=True):
        """ Return the newest version up to offset versions
        
        Args:
            name (str):the identifier of the object
            offset (int):the offset
            throw_error_not_exist (bool):true - throw error if not exists, else return []. Defaults to True.
        
        Raises:
            Exception: raises an exception if the object does not exists and throw_error_not_exist == True
        
        Returns:
            str -- the version
        """
        with closing(self._conn.cursor()) as cursor:
            if offset > 0:
                inner_select = "(select version, uuid_time from versions where name = '" + \
                    name + "' order by uuid_time ASC LIMIT " + str(offset) + ")"
                stmt = "select version from " + inner_select + " order by uuid_time DESC LIMIT 1"
                for row in cursor.execute(stmt):
                    result = row[0]
                    return result
            elif offset < 0:
                inner_select = "(select version, uuid_time from versions where name = '" + \
                    name + "' order by uuid_time DESC LIMIT " + str(-offset) + ")"
                stmt = "select version from " + inner_select + " order by uuid_time ASC LIMIT 1"
                for row in cursor.execute(stmt):
                    result = row[0]
                    return result
            if offset == 0:
                return self.get_first_version(name)
            if throw_error_not_exist:
                logger.error('No object with name ' + name + ' exists.')
                raise Exception('No object with name ' + name + ' exists.')
            else:
                return []

    def replace(self, obj):
        """ Overwrite existing object without incrementing version
        
        Args:
            obj (RepoObject): repo object to be overwritten
        """

        logger.info('Replacing ' + obj["repo_info"][RepoInfoKey.NAME.value] +
                    ', version ' + str(obj["repo_info"][RepoInfoKey.VERSION.value]))
        select_statement = "select path, file from versions where name = '" +\
            obj["repo_info"][RepoInfoKey.NAME.value] + "' and version = '" +\
            str(obj["repo_info"][RepoInfoKey.VERSION.value]) + "'"
        with closing(self._conn.cursor()) as cursor:
            for row in cursor.execute(select_statement):
                self._save_function(self._main_dir + '/' +
                                    str(row[0]) + '/' + str(row[1]), obj)
            # delete all modification infos
            cursor.execute("delete from modification_info where name='" + obj["repo_info"][RepoInfoKey.NAME.value] + "' and version = '" +
                    str(obj["repo_info"][RepoInfoKey.VERSION.value]) + "'")
            if repo_objects.RepoInfoKey.MODIFICATION_INFO.value in obj['repo_info']:
                for k, v in obj['repo_info'][repo_objects.RepoInfoKey.MODIFICATION_INFO.value].items():
                    tmp = _time_from_version(v)
                    cursor.execute("insert into modification_info (name, version, modifier, modifier_version, modifier_uuid_time) VALUES ('"
                            + obj["repo_info"][RepoInfoKey.NAME.value] + "','" + str(obj["repo_info"][RepoInfoKey.VERSION.value]) + "','" + k + "','" + str(v) + "','" + str(tmp) + "')")
            self._conn.commit()
          
    def close_connection(self):
        """ Closes the database connection
        """
        self._conn.close()

    def _get_all_files(self):
        """ Returns set of all files in directory
        
        Returns:
            set -- set with all filenames (including relative paths)
        """

        f = set()
        for path, subdirs, files in os.walk(self._main_dir):
            if '.git' in subdirs:
                subdirs.remove('.git')
            ignored_files = [f_ for f_ in files if f_.startswith('.')]
            for name in files:
                if not name in ignored_files:
                    #name = path + "/" + os.path.splitext(name)[0]
                    #name = name.replace('\\','/')
                    f.add(os.path.splitext(name)[0])
                    #path_to_file[name] = path
        return f

    def check_integrity(self):
        """ Checks if files are missing or have not yet been added
        
        Returns:
            dictionary -- contains sets of missing files and/or set of files not yet added
        """

        # first get list of all files
        f = self._get_all_files()
        repo_f = set()
        with closing(self._conn.cursor()) as cursor:
            for row in cursor.execute("select file from versions"):
                repo_f.add(row[0])
        tmp = f.copy()
        files_not_added = tmp-repo_f
        result = {}
        if(len(files_not_added) > 0):
            result = {'untracked files': files_not_added}
        files_missing = repo_f - f
        if(len(files_missing) > 0):
            result = {'deleted files': files_missing}
        return result

    def _cleanup(self):
        """ Checks the integrity of the files
        """
        
        files = self.check_integrity()
        print(files)


import pailab.version
from pailab.ml_repo.repo import MLRepo, MLObjectType
from pailab.ml_repo.repo_objects import RepoInfoKey, MeasureConfiguration, RawData, DataSet
from pailab.ml_repo.repo_objects import repo_object_init

from pailab.job_runner.job_runner import JobState

from pailab.ml_repo.repo_store import FIRST_VERSION, LAST_VERSION
